from Camera import Camera

from abc import ABC

import numpy as np

import cv2

from skimage import io

from io import BytesIO

from IPython.display import clear_output, Image, display, update_display



import PIL





try:

    from Cameras.PySpinCapture import PySpinCapture as psc

    print('1')

except ImportError:

    

    PySpinCapture = None



class Flir(Camera, ABC):

    def __init__(self, exposure=0.01, white_balance=1, auto_focus=False, grayscale=False):

        self._isMonochrome = True

        self._is16bits = True

        self.Cam = psc(0, self._isMonochrome, self._is16bits)

        fps = self.getFPS()

        resolution = self.getResolution()

        super().__init__(exposure, white_balance, auto_focus, fps, resolution, grayscale)

        self.hdr_exposures = None



    def getImage(self, name='test', saveImage=True, saveNumpy=True, calibration=False, timeout=5000):

        try:

            # Take and return current camera frame

            success, img = self.Cam.grabFrame()

            # Save if desired

            if saveImage:

                if calibration:

                    filenamePNG = 'CalibrationImages/' + name + '.PNG'

                    cv2.imwrite(filenamePNG, cv2.cvtColor(img, cv2.COLOR_BGR2RGB))

                else:

                    filename = 'CapturedImages/' + name + '.PNG'

                    cv2.imwrite(filename, cv2.cvtColor(img, cv2.COLOR_BGR2RGB))

            if saveNumpy:

                if calibration:

                    np.save('CalibrationNumpyData/' + name, img)

                else:

                    np.save('CapturedNumpyData/' + name, img)

            # self.Cam.release()

            return img

        except SystemError:

            self.quit_and_open()

            return None



    def setExposure(self, exposure):

        self.Cam.setExposure(exposure)

        



    def getExposure(self):

        return self.Cam.getExposure()



    def getFPS(self):

        return self.Cam.getFPS()



    def setFPS(self, fps):

        self.Cam.setFPS(fps)



    def setAutoGain(self):

        self.Cam.setCamAutoProperty()



    def getGain(self):

        return self.Cam.getGain()



    def setGain(self, gain):

        self.Cam.setGain(gain)



    def getResolution(self):

        size = self.Cam.getResolution()

        

        return size



    def setResolution(self, resolution):

        self.Cam.setWidth(resolution[0])

        self.Cam.setHeight(resolution[1])

        

    def getHDRImage(self, name='test', saveImage=True, saveNumpy=True):

        self.Cam.setupHDR()

        imgs = self.Cam.captureHDR()

        if saveNumpy:

            np.save('CapturedNumpyData/' + name, imgs)

        if saveImage:

            png_frame = (imgs - np.min(imgs)) / (np.max(imgs) - np.min(imgs))

            png_frame *= 255.0

            io.imsave('CapturedImages/' + name + '.PNG', png_frame.astype(np.uint8))



    def viewCameraStream(self):

        img = self.getImage(saveImage=False, saveNumpy=False)

        while True:

            _,img = self.Cam.grabFrameCont()

            cv2.imshow('FLIR camera image', img)

            c = cv2.waitKey(1)

            if c != -1:

                # When everything done, release the capture

                self.Cam._camera.EndAcquisition()

                cv2.destroyAllWindows()

                self.quit_and_open()

                break

            

    def viewCameraStreamJupyter(self):

        # Live view in a Jupyter Notebook

        try:

            start = self.getImage(saveImage=False, saveNumpy=False)

            g = BytesIO()

            PIL.Image.fromarray(start).save(g, 'jpeg')

            obj = Image(data=g.getvalue())

            dis = display(obj, display_id=True)

            while True:

                img = self.getImage(saveImage=False, saveNumpy=False)

                if img is None:

                    break

                img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)

                f = BytesIO()

                PIL.Image.fromarray(img).save(f, 'jpeg')

                obj = Image(data=f.getvalue())

                update_display(obj, display_id=dis.display_id)

                clear_output(wait=True)

        except KeyboardInterrupt:

            self.quit_and_open()

            

    def viewCameraStreamJupyterWindows(self):

        # Live view in a Jupyter Notebook

        try:

            start = self.getImage(saveImage=False, saveNumpy=False)

            g = BytesIO()

            PIL.Image.fromarray(start).save(g, 'jpeg')

            obj = Image(data=g.getvalue())

            dis = display(obj, display_id=True)

            while True:

                img = self.getImage(saveImage=False, saveNumpy=False)

                img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)

                f = BytesIO()

                PIL.Image.fromarray(img).save(f, 'jpeg')

                obj = Image(data=f.getvalue())

                update_display(obj, display_id=dis.display_id)

        except KeyboardInterrupt:

            self.quit_and_open()



    def quit_and_close(self):

        self.Cam.release()



    def quit_and_open(self):

        self.Cam.release()

        self.Cam = psc(1, self._isMonochrome, self._is16bits)



    def getStatus(self):

        raise NotImplementedError




import time
import threading
from queue import Queue
import cv2
import numpy as np

class imgShow:
    """
    Class that continuously shows a frame using a dedicated thread.
    """
    def __init__(self, displayHeight=1024, displayWidth=768, frame=None):
        self.frame = frame
        self.stopped = False
        self.displayHeight = displayHeight
        self.displayWidth = displayWidth
        self.thread = threading.Thread(target=self.show, args=())
        self.thread.daemon = True
        self.paused = True  # Start out paused.
        self.state = threading.Condition()

    # def __enter__(self):
    #     return self

    def start(self):
        self.thread.start()
        self.paused = False
        return self

    def show(self):

        while not self.stopped:
            with self.state:
                if self.paused:
                    self.state.wait()
            image_display = cv2.resize(self.frame, (self.displayHeight, self.displayWidth))
            cv2.imshow("Live View", image_display)
            if cv2.waitKey(1) == 27:
                self.paused = True
                self.stop()

    def resume(self):
        with self.state:
            self.paused = False
            self.state.notify()  # Unblock self if waiting.

    def pause(self):
        with self.state:
            self.paused = False
            self.state.notify()

    def stop(self):
        self.stopped = True
        # wait until stream resources are released (producer thread might be still grabbing frame)
        #self.thread.join()

    def __exit__(self):
        self.stop()


class ptgCamStream:
    def __init__(self, camera, transform=None, queue_size=128):
        # initialize the file video stream along with the boolean
        # used to indicate if the thread should be stopped or not
        self.stream = camera
        self.stopped = False
        self.transform = transform

        # initialize the queue used to store frames read from
        # the video file
        self.Q = Queue(maxsize=queue_size)
        # intialize thread
        self.thread = threading.Thread(target=self.update, args=())
        self.thread.daemon = True
        self.paused = True  # Start out paused.
        self.state = threading.Condition()
    # def __enter__(self):
    #     return self

    def start(self):
        # start a thread to read frames from the file video stream
        self.stream.setAcquisitMode(1)  # set to continuous acquisition
        self.stream.beginAcquisit(True)
        if self.stream._camera.IsStreaming():
            print("Camera Streaming")
        self.paused = False
        self.thread.start()
        return self

    def update(self):
        # keep looping infinitely
        while True:
            # if the thread indicator variable is set, stop the
            # thread
            if cv2.waitKey(1) == 27:
                self.paused = True
                self.stop()

            if self.stopped:
                break

            with self.state:
                if self.paused:
                    self.state.wait()

            # otherwise, ensure the queue has room in it
            if not self.Q.full():
                # read the next frame from the file
                #(grabbed, frame) = self.stream.read()

                # Single frame
                #grabbed, frame = self.stream.grabFrame()

                #Continuous
                grabbed, frame = self.stream.grabFrameCont()


                # if the `grabbed` boolean is `False`, then we have
                # reached the end of the video file
                if not grabbed:
                    self.stopped = True

                # if there are transforms to be done, might as well
                # do them on producer thread before handing back to
                # consumer thread. ie. Usually the producer is so far
                # ahead of consumer that we have time to spare.
                #
                # Python is not parallel but the transform operations
                # are usually OpenCV native so release the GIL.
                #
                # Really just trying to avoid spinning up additional
                # native threads and overheads of additional
                # producer/consumer queues since this one was generally
                # idle grabbing frames.
                if self.transform:
                    frame = self.transform(frame)

                # add the frame to the queue
                self.Q.put(frame)
            else:
                time.sleep(0.1)  # Rest for 10ms, we have a full queue

        # Continuous
        self.stream.beginAcquisit(False)
        #self.stream.release()

    def resume(self):
        with self.state:
            self.paused = False
            self.state.notify()  # Unblock self if waiting.

    def pause(self):
        with self.state:
            self.paused = False
            self.state.notify()

    def read(self):
        # return next frame in the queue
        return self.Q.get()

    # Insufficient to have consumer use while(more()) which does
    # not take into account if the producer has reached end of
    # file stream.
    def running(self):
        return self.more() or not self.stopped

    def more(self):
        # return True if there are still frames in the queue. If stream is not stopped, try to wait a moment
        tries = 0
        while self.Q.qsize() == 0 and not self.stopped and tries < 5:
            time.sleep(0.1)
            tries += 1

        return self.Q.qsize() > 0

    def stop(self):
        # indicate that the thread should be stopped
        self.stopped = True
        # wait until stream resources are released (producer thread might be still grabbing frame)
        #self.thread.join()

    def __exit__(self):
        self.stop()

class patternDisplay:
    """
    Class that continuously shows a frame using a dedicated thread.
    """
    def __init__(self, displayHeight=1080, displayWidth=1920, pattern=None):
        self.pattern = pattern
        self.stopped = False
        self.displayHeight = displayHeight
        self.displayWidth = displayWidth

        self.thread = threading.Thread(target=self.show, args=())
        self.thread.daemon = True

        print("0.3")
    # def __enter__(self):
    #     return self

    def start(self):
        self.thread.start()
        return self

    def show(self):
        print("0.4")
        while not self.stopped:
            window_name = 'projector'
            cv2.namedWindow(window_name, cv2.WND_PROP_FULLSCREEN)

            # If projector is placed right to main screen (see windows properties in your operating system)
            # if the pattern is displayed at the wrong monitor you need to play around with the coordinates here until the image is displayed at the right screen
            cv2.moveWindow(window_name, self.displayWidth, 0)

            cv2.setWindowProperty(window_name, cv2.WND_PROP_FULLSCREEN, cv2.WINDOW_FULLSCREEN)

            cv2.imshow(window_name, self.pattern.astype(np.float32))
            k = cv2.waitKey(0)
            if k == 27:  # wait for ESC key to exit
                self.stop()
                cv2.destroyAllWindows()

    def stop(self):
        self.stopped = True
        # wait until stream resources are released (producer thread might be still grabbing frame)
        #self.thread.join()

    def __exit__(self):
        self.stop()


from abc import ABC
from GhostScan.CalibrationSession import CalibrationSession
from GhostScan.Calibrations import RadiometricCalibration
import cv2
import os, sys, time, math
import numpy as np
import matplotlib.pyplot as plt
import os
DISPLAY_HEIGHT = 1080
DISPLAY_WIDTH = 1920

class RadiometricCalibDisplaySession(CalibrationSession, ABC):
    def __init__(self, camera, radio_calib, path='CalibrationImages/DisplayRadiometric', exposure=60000, numPhaseShift = 4):
        # Set camera, destination path
        self.camera = camera
        self.path = path
        # self.g = None
        self.radio_calib = radio_calib
        self.g = self.radio_calib.load_calibration_data()
        self.exposure = exposure
        self.camera.setExposure(exposure)
        self.NumPatterns = 256
        self.displayWidth = DISPLAY_WIDTH
        self.displayHeight = DISPLAY_HEIGHT
        self.setDefPattern()
        self.sequenceNum = 2 * numPhaseShift
        
        if not os.path.exists('CapturedImages/'):
            os.mkdir('CapturedImages/')
        if not os.path.exists('CapturedImages/sequenceImages/'):
            os.mkdir('CapturedImages/sequenceImages/')
        if not os.path.exists('CapturedImages/sequenceImages/undistortRadioCalib/'):
            os.mkdir('CapturedImages/sequenceImages/undistortRadioCalib/')
        if not os.path.exists('CapturedImages/sequenceImages/undistortRadioCalib/radioCalibResults/'):
            os.mkdir('CapturedImages/sequenceImages/undistortRadioCalib/radioCalibResults/')
        if not os.path.exists('CapturedImages/sequenceImages/undistortRadioCalib/undistortResults/'):
            os.mkdir('CapturedImages/sequenceImages/undistortRadioCalib/undistortResults/')
        if not os.path.exists('CalibrationImages/RadiometricTest/'):
            os.mkdir('CalibrationImages/RadiometricTest/')
        if not os.path.exists('CalibrationNumpyData/RadiometricTest/'):
            os.mkdir('CalibrationNumpyData/RadiometricTest/')
        if not os.path.exists('CalibrationImages/DisplayRadiometric'):
            os.mkdir('CalibrationImages/DisplayRadiometric')
            

        

    def calculateDisplayCalibration(self):
        
        pixelValues = self.avgImagePixel()

        DisplayToRadianceData = []

        for pixelValue in pixelValues:
            num = math.exp(self.g[round(pixelValue)] - math.log(self.exposure))
            DisplayToRadianceData.append(num)

        imgSave = 'CapturedImages/sequenceImages/undistortRadioCalib/radioCalibResults'
            
        fig = plt.figure()
        plt.title("Display to Relative Radiance Curve")
        plt.xlabel('Display pixel value')
        plt.ylabel('Relative Radiance')
        plt.plot(pixelValues, DisplayToRadianceData, 'k')
        # plt.show()
        fig.savefig(imgSave + '/Display to Relative Radiance Curve' + '.png')
        # fig.savefig('Display to Relative Radiance Curve.png')
        return DisplayToRadianceData, pixelValues

    def calUpValue(self, imgPath):
        img = cv2.imread(imgPath, cv2.IMREAD_GRAYSCALE)
        img = np.array(img, dtype='uint8')
        row, col = img.shape[0], img.shape[1]
        arr = []
        for i in range(row):
            cols = []
            for j in range(col):
                val = math.exp(self.g[img[i][j]] - math.log(self.exposure))
                cols.append(val)
            arr.append(cols)
        return np.array(arr)


    def calculateReflectivity(self):
        imgPath = 'CapturedImages/sequenceImages/RadioDisplayTest/0.png'
        arr = self.calUpValue(imgPath)

        # img = cv2.imread(imgPath, cv2.IMREAD_GRAYSCALE)
        # img = np.array(img, dtype='uint8')
        # row, col = img.shape[0], img.shape[1]

        DisplayToRadianceData, pixelValues = self.calculateDisplayCalibration()
        DisplayToRadianceValue = DisplayToRadianceData[255]
        
        # arr = []
        # for i in range(row):
        #     cols = []
        #     for j in range(col):
        #         val = math.exp(self.g[img[i][j]] - math.log(self.exposure)) / DisplayToRadianceValue
        #         cols.append(val)
        #     arr.append(cols)
        # arr = np.array(arr)
        # print(arr)
        
        arr = [[j/DisplayToRadianceValue for j in i] for i in arr]
        imgSave = 'CapturedImages/sequenceImages/undistortRadioCalib/undistortResults'
        fig = plt.figure()
        plt.imshow(arr, cmap='viridis')
        plt.colorbar()
        # plt.show()
        fig.savefig(imgSave + '/Reflectivity Map' + '.png')
        print('Reflectivity Calculation Successful! Reconstructed images viewable at CapturedImages/sequenceImages/undistortRadioCalib/undistortResults')
        # plt.savefig('Reflectivity Map.png')

        Reflectivity = np.array(arr)
        return Reflectivity, DisplayToRadianceData, pixelValues

    def radiometricCalibUndistortImages(self):
        Reflectivity = self.calculateReflectivity()
        imgSeqPath = 'CapturedImages/sequenceImages/RadioDisplayTest/'
        # sequenceNum = 8
        for i in range(self.sequenceNum):
            upValue = self.calUpValue(imgSeqPath + str(i) + '.png')
            # print(upValue)
            radiance = np.divide(upValue, Reflectivity)
            # print(radiance)
            # print(radiance.shape[0], radiance.shape[1])
            correctedImage = self.imageCorrection(radiance)
            # print(correctedImage)
            cv2.imwrite('CapturedImages/sequenceImages/undistortRadioCalib/undistortResults/' + str(i) + '.png', correctedImage)

    def imageCorrection(self, radiance):
        DisplayToRadianceData = self.calculateDisplayCalibration()
        row, col = radiance.shape[0], radiance.shape[1]
        imgCorr = []
        for i in range(row):
            cols = []
            for j in range(col):
                t = np.array(abs(DisplayToRadianceData - radiance[i][j]))
                t_min = np.min(t)
                tt = self.findPixelValue(t_min, t)
                cols.append(tt)
            imgCorr.append(cols)
        return np.array(imgCorr)

    def findPixelValue(self, value, arr):
        for i in range(len(arr)):
            if arr[i] == value:
                return i


    def setPatternScale(self):
        patternSize = min(self.displayHeight, self.displayWidth)
        boarderSize = math.floor((max(self.displayHeight, self.displayWidth) - patternSize)/2)

        self.patterns = np.zeros((self.displayHeight, self.displayWidth, self.NumPatterns), dtype=np.float32)
        
        return patternSize, boarderSize

    def setDefPattern(self):
        #patternSize, boarderSize = self.setPatternScale(2)
        patternSize, boarderSize = self.setPatternScale()

        patternSize = max(self.displayHeight, self.displayWidth)
        #patternSize = min(self.displayHeight, self.displayWidth)
        
        for i in range(self.NumPatterns):
            self.patterns[..., i] = i * np.ones((DISPLAY_HEIGHT,DISPLAY_WIDTH))


    def captureImages(self):
        cv2.destroyAllWindows()
        cv2.waitKey(1)
        cv2.waitKey(1)
        
        window_name = 'projector'
        cv2.namedWindow(window_name)
        cv2.moveWindow(window_name, self.displayWidth, 0)  #self.displayWidth
            
        
        if not os.path.exists('CalibrationNumpyData/'):
            os.mkdir('CalibrationNumpyData/')
        if not os.path.exists('CalibrationNumpyData/DisplayRadiometric/'):
            os.mkdir('CalibrationNumpyData/DisplayRadiometric/')
       
        for i in range(self.NumPatterns):
            if i % 10 == 0:
                print('picture taken:' + str(i))

            cur_img = self.patterns[... ,i]
            cv2.imshow(window_name, self.patterns[..., i].astype(np.uint8))
            cv2.waitKey(40)
            self.camera.getImage(name = 'DisplayRadiometric/'+str(i), calibration=True)
            
        print('Capture Complete! Numpy data viewable at CalibrationImages/DisplayRadiometric/')    
        cv2.waitKey(1000)
        cv2.destroyAllWindows()
        cv2.waitKey(1)
        cv2.waitKey(1)
        
    def captureTest(self):
        cv2.destroyAllWindows()
        cv2.waitKey(1)
        cv2.waitKey(1)
        
        window_name = 'projector'
        cv2.namedWindow(window_name)
        cv2.moveWindow(window_name, self.displayWidth, 0)  #self.displayWidth
            
        
        if not os.path.exists('CalibrationNumpyData/'):
            os.mkdir('CalibrationNumpyData/')
        if not os.path.exists('CalibrationImages/RadiometricTest/'):
            os.mkdir('CalibrationImages/RadiometricTest/')
       
        pattern = np.ones((DISPLAY_HEIGHT,DISPLAY_WIDTH))
        for i in range(DISPLAY_WIDTH):
            pattern[:,i] = i / DISPLAY_WIDTH * 255
        
        cv2.imshow(window_name, pattern.astype(np.uint8))
        cv2.waitKey(40)
        self.camera.getImage(name = 'RadiometricTest/'+ 'testImage', calibration=True)
        
        pattern = np.ones((DISPLAY_HEIGHT,DISPLAY_WIDTH)) * 255
        cv2.imshow(window_name, pattern.astype(np.uint8))
        cv2.waitKey(40)
        self.camera.getImage(name = 'RadiometricTest/'+ 'testImageDim', calibration=True)
        
        pattern = np.ones((DISPLAY_HEIGHT,DISPLAY_WIDTH)) * 128
        cv2.imshow(window_name, pattern.astype(np.uint8))
        cv2.waitKey(40)
        self.camera.getImage(name = 'RadiometricTest/'+ 'vignettingcalibration', calibration=True)
        
        print('Capture Complete! Test image viewable at CalibrationImages/RadiometricTest/')
        cv2.waitKey(1000)
        cv2.destroyAllWindows()
        cv2.waitKey(1)
        cv2.waitKey(1) 
        
        
        
    def avgImagePixel(self):
        pixelValues = []
        files = []
        # Load images
        
        for file in os.listdir(self.path):
            # Only use .raw files(Balser)
            #if file.endswith(".raw") or file.endswith(".Raw") or file.endswith(".RAW"):
            #    files.append(file)
            if file.endswith(".png") or file.endswith(".Png") or file.endswith(".PNG"):
                files.append(file)
        # Sort files depending on their exposure time from lowest to highest        
        files.sort(key=lambda x: int(x[:-4]))
        # We used exposure time as filenames

        for filename in files:
            #image = np.fromfile(self.path + '/' + filename, dtype=np.uint8)
            # image = cv2.imread(self.path + '/' + filename, cv2.IMREAD_UNCHANGED);
            image = cv2.imread(self.path + '/' + filename, 0)
            average = image.mean(axis=0).mean(axis=0) 
            pixelValues.append(average)

        x = list(range(256))
        imgSave = 'CapturedImages/sequenceImages/undistortRadioCalib/radioCalibResults'
        fig = plt.figure()
        plt.title("Display Value to Average Pixel Value Curve")
        plt.xlabel('Display pixel value')
        plt.ylabel('Average camera pixel value')
        plt.plot(x, pixelValues)
        # plt.show()
        fig.savefig(imgSave + '/Display Value to Average Pixel Value Curve' + '.png')
        # plt.savefig('Display Value to Average Pixel Value Curve.png')

        return pixelValues


    # def calibrate_HDR(self, smoothness=1000):
    #     # Call radiometric calibration
    #     # This computes and returns the camera response function and calculates a HDR image saved in path as PNG and .np
    #     self.g, le = self.radio_calib.get_camera_response(smoothness)
    #     self.radio_calib.plotCurve("Camera response")
    #     return self.g, le

    def load_calibration(self):
        self.radio_calib.load_calibration_data()

    def calibrate_image(self, exposure, path='CalibrationImages/Distorted'):
        # Calibrate radiometric calibrated images from a single exposure
        # Returns set of undistorted images and corresponding g function
        imgs, g = self.radio_calib.calibrate_image(exposure, path)
        return imgs, g





import os, math, time
import numpy as np
import cv2
import PySpin

class PySpinCapture:
    def __init__(self, index=0, isMonochrome=False, is16bits=False):
        # [Current Support] Single camera usage(select by index)
        self._system = PySpin.System.GetInstance()
        # Get current library version
        self._version = self._system.GetLibraryVersion()
        print('Library version: {}.{}.{}.{}\n'.format(self._version.major, self._version.minor, self._version.type,
                                self._version.build))
        self.index = index
        self.getNumCams()
        self._camera = self._cameraList.GetByIndex(index)
        self._camera.Init()
        self._isMonochrome = isMonochrome
        self._is16bits = is16bits

        self._nodemap = self._camera.GetNodeMap()
        self.getNode()

        self.setAcquisitMode(1)
        self.setPixel()
        self.setSize()
        self.setCamAutoProperty(False)
        #self._camera.BeginAcquisition()

    def __enter__(self):
        return self

    def reset(self):
        self.__init__()

    def getNumCams(self):
        self._cameraList = self._system.GetCameras()

    def print_retrieve_node_failure(self, node, name):

        print('Unable to get {} ({} {} retrieval failed.)'.format(node, name, node))
        print('The {} may not be available on all camera models...'.format(node))
        print('Please try a Blackfly S camera.')

    def getNode(self):
        # Acquisition Mode Node
        self.nodeAcquisitionMode = PySpin.CEnumerationPtr(self._nodemap.GetNode('AcquisitionMode'))
        if not PySpin.IsAvailable(self.nodeAcquisitionMode) or not PySpin.IsWritable(self.nodeAcquisitionMode):
            print("Unable to set acquisition mode to continuous (enum retrieval). Aborting...")
            return False

        # Retrieve entry node from enumeration node
        nodeContinuousAcquisition = self.nodeAcquisitionMode.GetEntryByName("Continuous")
        if not PySpin.IsAvailable(nodeContinuousAcquisition) or not PySpin.IsReadable(
                nodeContinuousAcquisition):
            print("Unable to set acquisition mode to continuous (entry retrieval). Aborting...")
            return False

        self.nodeAcquisitionContinuous = nodeContinuousAcquisition.GetValue()

        # Retrieve entry node from enumeration node
        nodeSingleAcquisition = self.nodeAcquisitionMode.GetEntryByName("SingleFrame")
        if not PySpin.IsAvailable(nodeSingleAcquisition) or not PySpin.IsReadable(
                nodeSingleAcquisition):
            print("Unable to set acquisition mode to Single Frame (entry retrieval). Aborting...")
            return False
        self.nodeAcquisitionSingle = nodeSingleAcquisition.GetValue()

        # Pixel Format Node
        self.nodePixelFormat = PySpin.CEnumerationPtr(self._nodemap.GetNode('PixelFormat'))

        nodePixelFormatMono8 = PySpin.CEnumEntryPtr(
            self.nodePixelFormat.GetEntryByName('Mono8'))
        self.pixelFormatMono8 = nodePixelFormatMono8.GetValue()

        nodePixelFormatMono16 = PySpin.CEnumEntryPtr(
            self.nodePixelFormat.GetEntryByName('Mono16'))
        self.pixelFormatMono16 = nodePixelFormatMono16.GetValue()

        # nodePixelFormatRGB8 = PySpin.CEnumEntryPtr(
        #     self.nodePixelFormat.GetEntryByName('BayerRG8'))
        # self.pixelFormatRGB8 = nodePixelFormatRGB8.GetValue()
        #
        # nodePixelFormatRGB16 = PySpin.CEnumEntryPtr(
        #     self.nodePixelFormat.GetEntryByName('BayerRG16'))
        # self.pixelFormatRGB16 = nodePixelFormatRGB16.GetValue()

        # Image Size Node
        self.nodeWidth = PySpin.CIntegerPtr(self._nodemap.GetNode('Width'))
        self.nodeHeight = PySpin.CIntegerPtr(self._nodemap.GetNode('Height'))

        # Exposure Node
        self._nodeExposureAuto = PySpin.CEnumerationPtr(self._nodemap.GetNode('ExposureAuto'))
        if not PySpin.IsAvailable(self._nodeExposureAuto) or not PySpin.IsWritable(self._nodeExposureAuto):
            self.print_retrieve_node_failure('node', 'ExposureAuto')
            return False

        self._exposureAutoOff = self._nodeExposureAuto.GetEntryByName('Off')
        if not PySpin.IsAvailable(self._exposureAutoOff) or not PySpin.IsReadable(self._exposureAutoOff):
            self.print_retrieve_node_failure('entry', 'ExposureAuto Off')
            return False

        if PySpin.IsAvailable(self._nodeExposureAuto) and PySpin.IsWritable(self._nodeExposureAuto):
            self._exposureAutoOn = self._nodeExposureAuto.GetEntryByName('Continuous')

        self._nodeExposure = PySpin.CFloatPtr(self._nodemap.GetNode('ExposureTime'))
        self._exposureMin = self._nodeExposure.GetMin()
        self._exposureMax = self._nodeExposure.GetMax()

        # Gain Node
        self._nodeGainAuto = PySpin.CEnumerationPtr(self._nodemap.GetNode('GainAuto'))
        if not PySpin.IsAvailable(self._nodeGainAuto) or not PySpin.IsWritable(self._nodeGainAuto):
            self.print_retrieve_node_failure('node', 'GainAuto')
            return False

        self._gainAutoOff = self._nodeGainAuto.GetEntryByName('Off')
        if not PySpin.IsAvailable(self._gainAutoOff) or not PySpin.IsReadable(self._gainAutoOff):
            self.print_retrieve_node_failure('entry', 'GainAuto Off')
            return False

        if PySpin.IsAvailable(self._nodeGainAuto) and PySpin.IsWritable(self._nodeGainAuto):
            self._gainAutoOn = self._nodeGainAuto.GetEntryByName('Continuous')

        self._nodeGain = PySpin.CFloatPtr(self._nodemap.GetNode('Gain'))
        self._gainMin = self._nodeGain.GetMin()
        self._gainMax = self._nodeGain.GetMax()

        self.nodeGammaEnable = PySpin.CBooleanPtr(self._nodemap.GetNode('GammaEnable'))
        if not PySpin.IsAvailable(self.nodeGammaEnable) or not PySpin.IsWritable(self.nodeGammaEnable):
            self.print_retrieve_node_failure('node', 'GammaEnable')
            return False

        nodeAcquisitionFrameRate = PySpin.CFloatPtr(self._nodemap.GetNode('AcquisitionFrameRate'))
        if not PySpin.IsAvailable(nodeAcquisitionFrameRate) or not PySpin.IsWritable(nodeAcquisitionFrameRate):
            print("Unable to set acquisition Frame Rate (enum retrieval). Aborting...")
            return False
        else:
            nodeAcquisitionFrameRate.SetValue(5)

            
    def setAcquisitMode(self, mode=0):
        if mode==0:
            #Single Frame mode
            self.nodeAcquisitionMode.SetIntValue(self.nodeAcquisitionSingle)
        elif mode==1:
            # Continuous mode
            self.nodeAcquisitionMode.SetIntValue(self.nodeAcquisitionContinuous)
            #self.nodeAcquisitionMode.SetValue(PySpin.AcquisitionMode_Continuous)

        #print(self.nodeAcquisitionMode.GetIntValue())
    
    
    def setFPS(self, fps):

        nodeAcquisitionFrameRate = PySpin.CFloatPtr(self._nodemap.GetNode('AcquisitionFrameRate'))

        if not PySpin.IsAvailable(nodeAcquisitionFrameRate) or not PySpin.IsWritable(nodeAcquisitionFrameRate):

            print("Unable to set acquisition Frame Rate (enum retrieval). Aborting...")

            return False

        else:

            nodeAcquisitionFrameRate.SetValue(fps)
    

    def getFPS(self):

        nodeAcquisitionFrameRate = PySpin.CFloatPtr(self._nodemap.GetNode('AcquisitionFrameRate'))

        return nodeAcquisitionFrameRate.GetValue()


    def setPixel(self):
        # Set the pixel format.
        if self._isMonochrome:
            # Enable Mono8 mode.
            if self._is16bits:
                self.nodePixelFormat.SetIntValue(self.pixelFormatMono16)
            else:
                self.nodePixelFormat.SetIntValue(self.pixelFormatMono8)
        # else:
        #     # Enable RGB8 mode.
        #     if self._is16bits:
        #         self.nodePixelFormat.SetIntValue(self.pixelFormatRGB16)
        #     else:
        #         self.nodePixelFormat.SetIntValue(self.pixelFormatRGB8)

    def setSize(self):
        self.width = self.nodeWidth.GetMax()
        self.nodeWidth.SetValue(self.width)

        self.height = self.nodeHeight.GetMax()
        self.nodeHeight.SetValue(self.height)

    def setCamAutoProperty(self, switch=True):
        # [Current Support] Gain, Exposure time
        # In order to manual set value, turn off auto first
        if switch:
            if PySpin.IsAvailable(self._exposureAutoOn) and PySpin.IsReadable(self._exposureAutoOn):
                self._nodeExposureAuto.SetIntValue(self._exposureAutoOn.GetValue())
                print('Turning automatic exposure back on...')

            if PySpin.IsAvailable(self._gainAutoOn) and PySpin.IsReadable(self._gainAutoOn):
                self._nodeGainAuto.SetIntValue(self._gainAutoOn.GetValue())
                print('Turning automatic gain mode back on...\n')

            self.nodeGammaEnable.SetValue(True)
        else:
            self._nodeExposureAuto.SetIntValue(self._exposureAutoOff.GetValue())
            print('Automatic exposure disabled...')

            self._nodeGainAuto.SetIntValue(self._gainAutoOff.GetValue())
            print('Automatic gain disabled...')

            #self.nodeGammaEnable.SetValue(False)
            #print('Gamma disabled...')
    
    def getSize(self):
        width = self.width
        height = self.height
        return width, height

    def getResolution(self):
        width,height = self.getSize()
        return (width,height)

    def setGain(self, gain_to_set):

        if float(gain_to_set) < self._gainMin or float(gain_to_set) > self._gainMax:
            print("[WARNING]: Gain value should be within {} to {}.(Input:{}) Set to half."
                  .format(self._gainMin, self._gainMax, float(gain_to_set)))
            self._nodeGain.SetValue(math.floor(self._gainMax + self._gainMin))
            print("Gain: {}".format(self._nodeGain.GetValue()))
        else:
            self._nodeGain.SetValue(float(gain_to_set))
            print("Gain: {}".format(self._nodeGain.GetValue()))

    def setExposure(self, exposure_to_set):

        exposure_to_set = float(exposure_to_set)
        if exposure_to_set < self._exposureMin or exposure_to_set > self._exposureMax:
            print("[WARNING]: Gain value should be within {} to {}.(Input:{}) Set to half."
                  .format(self._exposureMin, self._exposureMax, exposure_to_set))
            self._nodeExposure.SetValue(math.floor(self._exposureMax + self._exposureMin))
            print("Exposure: {}".format(self._nodeExposure.GetValue()))
        else:
            self._nodeExposure.SetValue(exposure_to_set)
            print("Exposure: {}".format(self._nodeExposure.GetValue()))

    def grabFrame(self):
        self.setAcquisitMode(0)
        self._camera.BeginAcquisition()
        cameraBuff = self._camera.GetNextImage()
        if cameraBuff.IsIncomplete():
            return False, None

        cameraImg = cameraBuff.GetData().reshape(self.height, self.width)
        image = cameraImg.copy()
        cameraBuff.Release()
        self._camera.EndAcquisition()

        if self._isMonochrome:
            return True, image
        else:
            image_rgb = cv2.cvtColor(image, cv2.COLOR_BayerRG2BGR)
            return True, image_rgb

    def grabFrameCont(self):
        cameraBuff = self._camera.GetNextImage()
        if cameraBuff.IsIncomplete():
            return False, None

        cameraImg = cameraBuff.GetData().reshape(self.height, self.width)
        image = cameraImg.copy()
        cameraBuff.Release()
        #self._camera.EndAcquisition()

        if self._isMonochrome:
            return True, image
        else:
            image_rgb = cv2.cvtColor(image, cv2.COLOR_BayerRG2BGR)
            return True, image_rgb

    def beginAcquisit(self, switch=True):
        if switch:
            self._camera.BeginAcquisition()
        else:
            self._camera.EndAcquisition()

    def getNextImage(self):
        return self._camera.GetNextImage()

    def setupHDR(self):

        if PySpin.IsAvailable(self._exposureAutoOn) and PySpin.IsReadable(self._exposureAutoOn):
            self._nodeExposureAuto.SetIntValue(self._exposureAutoOn.GetValue())
            print('Turning automatic exposure back on...')

        time.sleep(0.5)
        series = [2 ** (-2), 2 ** (-1), 1, 2 ** 1, 2 ** 2]
        midExposure = self._nodeExposure.GetValue()
        print("midExposure: ", midExposure)
        self.exposureHDRList = [midExposure * x for x in series]
        if self.exposureHDRList[0] < self._exposureMin:
            self.exposureHDRList[0] = self._exposureMin
        if self.exposureHDRList[-1] > self._exposureMax:
            self.exposureHDRList[-1] = self._exposureMax

        print("HDR Exposure List: ", self.exposureHDRList)

        self._nodeExposureAuto.SetIntValue(self._exposureAutoOff.GetValue())
        print('Automatic exposure disabled...')

    def captureHDR(self):
        if not hasattr(self, 'exposureHDRList'):
            print("[ERROR]: Need to setup HDR Exposure list first!!!")
            return 0
        imgs = np.zeros((len(self.exposureHDRList), self.height, self.width))
        for index, x in enumerate(self.exposureHDRList):
            self.setExposure(x)
            flag, tmp = self.grabFrame()
            if flag:
                imgs[index, ...] = tmp
            else:
                print("[WARNING]: Invalid Capture!!!")

        return imgs

    def release(self):

        # Turn auto gain and exposure back on in order to return the camera to tis default state
        self.setCamAutoProperty(True)
        if self._camera.IsStreaming():
            self._camera.EndAcquisition()
        self._camera.DeInit()
        #del self._camera
        # self._cameraList.Clear()
        # self._system.ReleaseInstance()

    def __exit__(self):
        self.release()



from abc import abstractmethod


class ImageProcessing:
    def __init__(self, capture_path):
        # Path to first dataset
        self.path = capture_path

    @abstractmethod
    def loadData(self):
        # To override: Load captured data
        pass

    @abstractmethod
    def computePhaseMaps(self):
        # To override: If applicable
        pass

    @abstractmethod
    def computeNormalMap(self, exposure):
        # To override: Compute normals
        pass

    @abstractmethod
    def computeDepthMap(self, exposure):
        # To override: Compute normals
        pass

    @abstractmethod
    def computePointCloud(self):
        # To override: Compute depth map - point cloud
        pass

    @abstractmethod
    def highPassFilter(self):
        # To override
        pass






from abc import abstractmethod


class CaptureSession:
    def __init__(self, camera, projection, calibration, image_processing):
        # Initialize your imaging set-up
        self.camera = camera
        self.projection = projection
        self.calibration = None
        self.image_processing = image_processing

    @abstractmethod
    def capture(self):
        # To override: Capture procedure
        pass

    @abstractmethod
    def compute(self, calibration):
        # To override: Compute procedure
        pass



from __future__ import print_function
import os, sys, glob
import numpy as np
import cv2
import scipy.spatial
import matplotlib
import wavepy
import matplotlib.pyplot as plt
# import pyvista as pv
# import plotly.graph_objects as go

# Set Parameters
DEFLECTOMETRY_GRADMAP_FILTERSIZE = 49

# Render parameters
MESH_Ka = [0.0, 0.0, 0.0]
MESH_Kd = [0.9, 0.9, 0.9]
MESH_Ks = [0.1, 0.1, 0.1]
MESH_Ns = 5.000000
MESH_d = 0.010000
MESH_illum = 1

# Scaling factors
MESH_DEPTH_SCALINGFACTOR = 5
MESH_SIGMA_GLOBCORR = 5


class Mesh:
    def __init__(self, name, height, width, cropMask=0):

        self.name = name
        self.height = height
        self.width = width
        self.vertex = np.zeros((self.height * self.width, 3))
        # set 2d plane
        self.setVertex2D()
        self.setFace()
        if cropMask == 0:
            self.mask = np.array([0, 0, self.height, self.width])
        else:
            self.mask = cropMask

    def setVertex2D(self):
        [i, j] = np.meshgrid(np.r_[1:self.height + 1], np.r_[1:self.width + 1])
        self.vertex2D = np.concatenate((i.flatten('F')[:, None], j.flatten('F')[:, None]), axis=1).astype(np.int16)
        self.vertex[:, 0:2] = self.vertex2D

    # options = 'Qt QbB Qc'
    # self.faces = scipy.spatial.Delaunay(v, qhull_options=options).simplices + 1

    def setFace(self):
        # N = 2
        options = 'Qt QbB Qc'
        self.faces = scipy.spatial.Delaunay(self.vertex2D, qhull_options=options).simplices + 1

    def setDepth(self):
        # Set x and y direction
        p = self.normals[..., 0] / self.normals[..., 2]
        q = self.normals[..., 1] / self.normals[..., 2]
        p = np.nan_to_num(p)
        q = np.nan_to_num(q)
        # TODO: Low pass filter to filter out the noise (BilateralFilter)
        # Compute depth as a imaginary number (real: depth, img: noise)
        # Using Frankot Chellappa method with FFT
        #x = sfg.frankotchellappa(p, q, False)
        x = wavepy.surface_from_grad.frankotchellappa(p, q, False)
        depth = x.real
        # depth_smooth = cv2.GaussianBlur(depth, (sigma_GlobCorr, sigma_GlobCorr), ((sigma_GlobCorr + 1) / 2))
        depth *= MESH_DEPTH_SCALINGFACTOR
        # cv.bilateralFilter(src, dst, 9, 75, 75, cv.BORDER_DEFAULT)
        depth_corr = cv2.bilateralFilter(depth.astype(np.float32), MESH_SIGMA_GLOBCORR, 90, 90)
        # depth_corr = depth
        # cv2.imwrite('./depth_before.png', (depth*255).astype(np.uint8))
        # depth = cv2.blur(depth.astype(np.float32), (5,5))
        # cv2.imwrite('./depth_after.png', (depth*255).astype(np.uint8))
        depth_crop = depth_corr[self.mask[0]:self.mask[2], self.mask[1]:self.mask[3]]
        self.vertex[:, 2] = depth_crop.reshape(-1)
        return depth

    def setNormal(self, normals):
        self.normals = normals

    def setTexture(self, texture):
        #self.texture = (cv2.flip(texture, -1) * 255).astype(np.uint8)
        self.texture = cv2.flip(texture, -1)


    # @nb.jit(nopython=True)
    def exportOBJ(self, path, withTexture=True):
        # totalCount = self.vertex.shape[0] + self.height * self.width + self.faces.shape[0]
        # progress = ProgressBar(totalCount, fmt=ProgressBar.FULL)
        print('Exporting OBJ...')
        N = self.normals[self.mask[0]:self.mask[2], self.mask[1]:self.mask[3]].reshape(-1, 3)
        filename = os.path.join(os.path.normpath(path), "mesh_" + self.name + ".obj")
        f = open(filename, 'w')
        f.write('mtllib material_' + self.name + '.mtl\n')
        f.write('usemtl Textured\n')
        for i in range(self.vertex.shape[0]):
            # progress.current += 1
            # progress()
            f.write('v %f %f %f\n' % (self.vertex[i, 0], self.vertex[i, 1], self.vertex[i, 2]))
            f.write('vn %f %f %f\n' % (N[i, 0], N[i, 1], N[i, 2]))

        if withTexture:
            for i in range(self.height):
                for j in range(self.width):
                    # progress.current += 1
                    # progress()
                    u = (i) / self.height
                    v = 1 - (j) / self.width
                    f.write('vt %f %f\n' % (v, u))

        for i in range(self.faces.shape[0]):
            # progress.current += 1
            # progress()
            # f.write('f ' + str(self.faces[i, 0]) + ' ' + str(self.faces[i, 1]) + ' ' + str(self.faces[i, 2]) + '\n')
            f.write(
                'f ' + str(self.faces[i, 0]) + '/' + str(self.faces[i, 0]) + '/' + str(self.faces[i, 0]) + ' ' + str(
                    self.faces[i, 1]) + '/' + str(self.faces[i, 1]) + '/' + str(self.faces[i, 1]) + ' ' + str(
                    self.faces[i, 2]) + '/' + str(self.faces[i, 2]) + '/' + str(self.faces[i, 2]) + '\n')

        f.close()

        if withTexture:
            # export mtl
            filename = os.path.join(os.path.normpath(path), "material_" + self.name + ".mtl")
            f = open(filename, 'w')
            f.write('newmtl Textured\n')
            f.write('Ka %f %f %f\n' % (MESH_Ka[0], MESH_Ka[1], MESH_Ka[2]))
            f.write('Kd %f %f %f\n' % (MESH_Kd[0], MESH_Kd[1], MESH_Kd[2]))
            f.write('Ks %f %f %f\n' % (MESH_Ks[0], MESH_Ks[1], MESH_Ks[2]))
            f.write('Ns %f\n' % (MESH_Ns))
            f.write('d %f\n' % (MESH_d))
            f.write('illum %d\n' % (MESH_illum))
            f.write("map_Kd texture_" + self.name + ".jpg\n")
            f.close()

            # export texture
            filename = os.path.join(os.path.normpath(path), "texture_" + self.name + ".jpg")
            # cv2.imwrite(filename, self.texture)
            # cv2.imwrite(filename, np.array((utilities.imgBrighten(self.texture, 10) * 255), dtype=np.uint8))
            cv2.imwrite(filename, self.texture)




from __future__ import print_function

import sys, os, math, glob, argparse
import numpy as np
import cv2
from cv2 import aruco
from checkerboard import detect_checkerboard
from tqdm import tqdm, trange
import matplotlib.pyplot as plt

PNGPattern = "*.png"

## Setup

# Checker Pattern(horizontal square count -1, vertical square count -1) project on display
CHECKERPATTERN_SIZE = (23, 7)
# Display dimensions
SCREEN_WIDTH = 1920
SCREEN_HEIGHT = 1080
# Checker Pattern coordination in pixel unit
CHECKER_FILE = "Projections/8_24_checker.npz"
# Aruco marker on the mirror dimensions (m)
ARUCO_MARKER_LENGTH = 0.04
#ARUCO_BOARD_HEIGHT = 0.2884
#ARUCO_BOARD_WIDTH = 0.5932
aruco_dict = aruco.Dictionary_get(aruco.DICT_6X6_250)

def readFileList(imgFolder, ImgPattern=PNGPattern):
    imgFileList = glob.glob(os.path.join(imgFolder, ImgPattern))
    imgFileList.sort()
    return imgFileList

def detectChecker(img, patternSize=CHECKERPATTERN_SIZE, debug=True):
    # print(len(img.shape))
    if len(img.shape) == 3:
        gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
    elif len(img.shape) == 2:
        gray = img
    #gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)

    ret, corners = cv2.findChessboardCorners(gray, patternSize, cv2.CALIB_CB_ADAPTIVE_THRESH + cv2.CALIB_CB_EXHAUSTIVE)
    corners_refine = corners
    #print(ret)
 
    if ret:
        criteria = (cv2.TERM_CRITERIA_EPS + cv2.TERM_CRITERIA_MAX_ITER, 30, 0.001)
        corners_refine = cv2.cornerSubPix(gray, corners, (3, 3), (-1, -1), criteria)

        if debug:
            cv2.drawChessboardCorners(img, patternSize, corners_refine, ret)
            width = 960
            height = int(img.shape[0] * 960 / img.shape[1])
            smallimg = cv2.resize(img, (width, height))
            # fig = plt.figure(figsize=(15,10))
            # plt.figure()
            # fig.imshow(smallimg)
            # print("image of checkerboard!!!!!!!!!!!!!!!!!!")
            # plt.show()
            # fig.show()

    return ret, corners_refine
def readCheckerObjPoint(fname):
    data = np.load(fname)
    objp = data["objp"]

    return objp
def arucoBoard(m=ARUCO_MARKER_LENGTH, h= 0.2884, w= 0.5932):

    # create objPoints for calibration target
    h0 = (0 - h / 2)
    hm = (m - h / 2)
    h1 = (((h - m) / 2) - h / 2)
    h2 = (((h + m) / 2) - h / 2)
    h3 = ((h - m) - h / 2)
    h4 = (h - h / 2)
    w0 = (0 - w / 2)
    wm = (m - w / 2)
    w1 = (((w - m) / 2) - w / 2)
    w2 = (((w + m) / 2) - w / 2)
    w3 = ((w - m) - w / 2)
    w4 = (w - w / 2)


    objPoints = []
    objPoints.append(np.array([[w0, h0, 0], [wm, h0, 0], [wm, hm, 0], [w0, hm, 0]], dtype=np.float32))  # 0
    objPoints.append(np.array([[w0, h1, 0], [wm, h1, 0], [wm, h2, 0], [w0, h2, 0]], dtype=np.float32))  # 1
    objPoints.append(np.array([[w0, h3, 0], [wm, h3, 0], [wm, h4, 0], [w0, h4, 0]], dtype=np.float32))  # 2
    objPoints.append(np.array([[w1, h3, 0], [w2, h3, 0], [w2, h4, 0], [w1, h4, 0]], dtype=np.float32))  # 3
    objPoints.append(np.array([[w3, h3, 0], [w4, h3, 0], [w4, h4, 0], [w3, h4, 0]], dtype=np.float32))  # 4
    objPoints.append(np.array([[w3, h1, 0], [w4, h1, 0], [w4, h2, 0], [w3, h2, 0]], dtype=np.float32))  # 5
    objPoints.append(np.array([[w3, h0, 0], [w4, h0, 0], [w4, hm, 0], [w3, hm, 0]], dtype=np.float32))  # 6
    objPoints.append(np.array([[w1, h0, 0], [w2, h0, 0], [w2, hm, 0], [w1, hm, 0]], dtype=np.float32))  # 7

    ids = np.linspace(0, 7, 8).astype(np.int32)[:, None]
    arucoCornerBoard = aruco.Board_create(objPoints, aruco_dict, ids)

    return arucoCornerBoard, objPoints

def detectAruco(img, debug=True):
    if len(img.shape) == 3:
        gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
    elif len(img.shape) == 2:
        gray = img

    parameters = aruco.DetectorParameters_create()
    #corners, ids, rejectedImgPoints = aruco.detectMarkers(img, aruco_dict, parameters=parameters)
    corners, ids, rejectedImgPoints = aruco.detectMarkers(gray, aruco_dict, parameters=parameters)

    if debug:
        frame_markers = aruco.drawDetectedMarkers(img.copy(), corners, ids)
        width = 960
        height = int(img.shape[0] * 960 / img.shape[1])
        smallimg = cv2.resize(frame_markers, (width, height))
        plt.figure(figsize=(15,10))
        plt.imshow(smallimg)
        plt.show()
        # print("image of Aruco!!!!!!!!!!!!!!!!!!")

    return corners, ids

def postEst(corners, ids, camMat, distCoeffs, markerLength=ARUCO_MARKER_LENGTH, arucoBoardHeight= 0.2884, arucoBoardWidth= 0.5932):

    arucoCornerBoard, _ = arucoBoard(markerLength, arucoBoardHeight, arucoBoardWidth)

    rvec = []
    tvec = []
    retval, rvec, tvec = aruco.estimatePoseBoard(corners, ids, arucoCornerBoard, camMat, distCoeffs, None, None)

    return rvec, tvec

def reProjAruco(img, camMat, distCoeffs, rvec, tvec, cornersAruco, markerLength=ARUCO_MARKER_LENGTH, arucoBoardHeight= 0.2884, arucoBoardWidth= 0.5932):
    print("reProjAruco")
    _, objPoints = arucoBoard(markerLength, arucoBoardHeight, arucoBoardWidth) #yunhao

    ids = np.linspace(0, 7, 8).astype(np.int32)[:, None]
    corners_reproj = []
    for i in range(len(objPoints)):
        imgPoints, _ = cv2.projectPoints(np.array(objPoints[i]), rvec, tvec, camMat, distCoeffs)
        corners_reproj.append(imgPoints)

    frame_markers = aruco.drawDetectedMarkers(img.copy(), corners_reproj, ids)
    # cv2.imwrite("./reproejct_markers.png", frame_markers)
    fig = plt.figure(figsize=(15,10))
    plt.imshow(frame_markers)
    imgPath = os.path.join(os.path.join(os.path.join(os.getcwd(), 'CalibrationImages'), 'Geometric'), 'geoCalibResults')
    fig.savefig(imgPath + '/reproejctMarkers' + '.png')
    plt.show()

def householderTransform(n, d):
    I3 = np.identity(3, dtype=np.float32)
    e = np.array([0, 0, 1])
    p1 = I3 - 2 * np.outer(n, n)
    p2 = I3 - 2 * np.outer(e, e)
    p3 = 2 * d * n

    return p1, p2, p3

def invTransformation(R, t):
    Rinv = R.T
    Tinv = -(Rinv@t)

    return Rinv, Tinv

def calib(imgPath, camMtx, dist, half_length, half_height, displayScaleFactor):
    ARUCO_BOARD_HEIGHT = half_height * 2 
    ARUCO_BOARD_WIDTH = half_length * 2 
    CHECKSQUARE_SIZE = displayScaleFactor

    objP_pixel = np.ceil(readCheckerObjPoint(CHECKER_FILE))
    objP_pixel[:, 2] = 0
    objP = np.array(objP_pixel)
    for i in range(CHECKERPATTERN_SIZE[1]):
        for j in range(math.floor(CHECKERPATTERN_SIZE[0]/2)):
            tmp = objP[CHECKERPATTERN_SIZE[0] * i + j, 0]
            objP[CHECKERPATTERN_SIZE[0] * i + j, 0] = objP[CHECKERPATTERN_SIZE[0] * i + CHECKERPATTERN_SIZE[0] - j - 1, 0]
            objP[CHECKERPATTERN_SIZE[0] * i + CHECKERPATTERN_SIZE[0] - j - 1, 0] = tmp
    objP[:, 0] -= (SCREEN_WIDTH / 2 - 1)
    objP[:, 1] -= (SCREEN_HEIGHT / 2 - 1)

    objP *= CHECKSQUARE_SIZE
    

    rtA = []
    rB = []
    tB = []
    rC2Ss = []
    tC2Ss = []

    # define valid image
    validImg = -1
    #for i in trange(len(imgFileList), desc="Images"):
    for i in range(1):
        
        #img = cv2.imread(imgPath, cv2.IMREAD_UNCHANGED)
        img = cv2.imread(imgPath)
        # print(img.shape)
        # Yunhao
        # img = (img/65535*255).astype(np.uint8)
        #print(len(img.shape))
        # plt.imshow(img)
        # plt.colorbar()
        # Aruco marker for Mirror position
        cornersAruco, ids = detectAruco(img, debug=False)
        if cornersAruco is None and ids is None and len(cornersAruco) <= 3:
            continue

        # Checker for Display
        # print(len(img.shape))
        ret, cornersChecker = detectChecker(img, CHECKERPATTERN_SIZE, debug=True)
        # print(ret)
        # print(cornersChecker)
        if not ret:
            print("no Checker!!!")
            continue
        
        #for a valid image, aruco and checker must be both detected
        validImg += 1

        # Calibrate Mirror Pose with Aruco

        rvecMirror, tvecMirror = postEst(cornersAruco, ids, camMtx, dist, ARUCO_MARKER_LENGTH,
                                         ARUCO_BOARD_HEIGHT, ARUCO_BOARD_WIDTH)
        img_axis = aruco.drawAxis(img, camMtx, dist, rvecMirror, tvecMirror, ARUCO_MARKER_LENGTH)
        width = 960
        height = int(img_axis.shape[0] * 960 / img_axis.shape[1])
        smallimg = cv2.resize(img_axis, (width, height))
        # plt.figure(figsize=(15,10))
        # plt.imshow(smallimg)
        # plt.show()


        ## Reproejct Camera Extrinsic
        reProjAruco(img, camMtx, dist, rvecMirror, tvecMirror, cornersAruco)

        rMatMirror, _ = cv2.Rodrigues(rvecMirror)  # rotation vector to rotation matrix
        normalMirror = rMatMirror[:, 2]

        rC2W, tC2W = invTransformation(rMatMirror, tvecMirror)
        dW2C = abs(np.dot(normalMirror, tvecMirror))

        # Householder transformation
        p1, p2, p3 = householderTransform(normalMirror, dW2C)

        # Calibrate virtual to Camera with Checker
        rpe, rvecVirtual, tvecVirtual = cv2.solvePnP(objP, cornersChecker, camMtx, dist, flags=cv2.SOLVEPNP_IPPE)  # cv2.SOLVEPNP_IPPE for 4 point solution #cv2.SOLVEPNP_ITERATIVE
                                                                      #iterationsCount=200, reprojectionError=8.0,

        rvecVirtual, tvecVirtual = cv2.solvePnPRefineLM(objP, cornersChecker, camMtx, dist, rvecVirtual,
                                                        tvecVirtual)

        proj, jac = cv2.projectPoints(objP, rvecVirtual, tvecVirtual, camMtx, dist)
        img_rep = img

        cv2.drawChessboardCorners(img_rep, CHECKERPATTERN_SIZE, proj, True)
        width = 960
        height = int(img_rep.shape[0] * 960 / img_rep.shape[1])
        smallimg = cv2.resize(img_rep, (width, height))
        fig = plt.figure(figsize=(15,10))
        plt.imshow(smallimg)
        imgPath = os.path.join(os.path.join(os.path.join(os.getcwd(), 'CalibrationImages'), 'Geometric'), 'geoCalibResults')
        fig.savefig(imgPath + '/chesssBoard' + '.png')
        plt.show()


        rMatVirtual, _ = cv2.Rodrigues(rvecVirtual)  # rotation vector to rotation matrix

        print(tvecVirtual)
        if validImg == 0:
            rtA = p1
            rB = np.matmul(rMatVirtual, p2)
            tB = np.squeeze(tvecVirtual) + p3
        else:
            rtA = np.concatenate((rtA, p1))
            rB = np.concatenate((rB, np.matmul(rMatVirtual, p2)))
            tB = np.concatenate((tB, np.squeeze(tvecVirtual) + p3))

        rS2C = p1 @ rMatVirtual
        tS2C = p1 @ np.squeeze(tvecVirtual) + p3

        rC2S, tC2S = invTransformation(rS2C, tS2C)
        print("rC2S:", rC2S)
        print("tC2S:", tC2S)
        rC2Ss.append(rC2S)
        tC2Ss.append(tC2S)

    return rC2Ss, tC2Ss


if __name__ == "__main__":

    parser = argparse.ArgumentParser()
    parser.add_argument("-d", "--directory", type=str, help="Path of images directory")
    parser.add_argument("-c", "--calibration", type=str, help="Path to intrinsic calibration npz")
    args = parser.parse_args()

    if args.directory is not None:
    	imgFolder = os.path.normpath(args.directory)
    if args.calibration is not None:
    	cameraCalibPath = os.path.normpath(args.calibration)

    calib(imgFolder, cameraCalibPath)


import numpy as np
import os
import cv2
import glob
import matplotlib.pyplot as plt
import skimage
from skimage.util import img_as_ubyte
from skimage.color import rgb2gray
import PIL

class RadiometricCalibration:

    def __init__(self, resolution, gamma=0.57, sampling_points=1000, path='CalibrationImages/Radiometric'):
        # # Set camera, destination path
        self.path = path
        # Get image resolution:
        self.width, self.height = resolution
        # print((self.width, self.height))
        # (1920, 1200)
        # Amount of sample points per image - to speed up calculation
        self.sampling_points = sampling_points
        # Initialize g function with None for later same for log exposure values
        # Gamma correction:
        self.gamma = gamma
        self.g = None
        self.w = None
        self.le = None
        # Raw captured data
        self.raw_data = None
        # Down-sampled data
        self.raw_samples = None
        # Exposure times
        self.exposures = None

    def load_calibration_data(self):
        # Check if calibration file already exists
        if os.path.exists('CalibrationNumpyData/radiometric.npz'):
            # Load g function, log exposure, weighting function, exposures, raw samples
            data = np.load('CalibrationNumpyData/radiometric.npz')
            self.g = data['g_function']
            self.le = data['log_exposures']
            self.w = data['w_function']
            self.exposures = data['exposures']
            self.raw_samples = data['samples']
        else:
            print("Capture and calibrate camera first")
        return self.g

    def compute_gamma_colorchart(self, intensities):
        # Intensities is an 1D-array of the capture intensity values of the gray tiles on the checker board
        # Returns a gamma values that fits the captured intensities to a linear plot
        no_intensities = intensities.shape[0]
        intensities = (intensities-np.min(intensities))/(np.max(intensities)-np.min(intensities))
        ground_truth = np.linspace(0, 1, no_intensities)
        # Disregard zero values because their logarithm is not defined
        self.gamma = np.sum(np.log(intensities[1:])*np.log(ground_truth[1:]))/np.sum(np.log(intensities[1:])**2)
        return self.gamma

    def load_raw_data(self):
        # Loading raw data files
        k = 0
        # Empty lists for images and exposure times
        Exposure = []
        self.raw_data = []
        files = []
        for file in os.listdir(self.path):
            # Only use .png files
            if file.endswith(".PNG") or file.endswith(".png") or file.endswith(".Png"):
                files.append(file)
        # Sort files depending on their exposure time from lowest to highest
        files.sort(key=lambda x: int(x[:-4]))
        # We used exposure time as filenames
        print("loading data..")
        for filename in files:
            image = PIL.Image.open(self.path + '/' + filename)
            image = np.asarray(image,dtype=np.uint16)
            image = rgb2gray(image)
            image = image * 65535
            print("Image" + str(k) + " intensity max:" + str(image.max()))
            # for .raw file, we need to know the picture shape in advance
            self.width = image.shape[0]
            self.height = image.shape[1]
            filename = os.path.splitext(filename)[0] + '\n'
            Exposure.append(int(filename))
            self.raw_data.append(image)
            # k is used to count the number of pictures
            k = k + 1

        ExposureNumber = k

        # Shrink the number of samples #
        Z = np.zeros([self.sampling_points, ExposureNumber], int)
        # Choose random sample points within image
        row = np.random.randint(self.width, size=self.sampling_points)
        col = np.random.randint(self.height, size=self.sampling_points)
        for i in range(ExposureNumber):
            Z[:, i] = self.raw_data[i][row, col]
        # Initialize to raw_samples
        self.raw_samples = Z
        exps = np.sort(np.array(Exposure))
        # Check if loaded exposure values match the predefined values
        self.exposures = exps
        print("Radiometric raw data loaded...")

    def plotCurve(self, title):
        """
        This function will plot the curve of the solved G function and the measured pixels. You don't need to return anything in this function.
        Input
        solveG: A (256,1) array. Solved G function generated in the previous section.
        LE: Log Erradiance of the image.
        logexpTime: (k,) array, k is the number of input images. Log exposure time.
        zValues: m*n array. m is the number of sampling points, and n is the number of input images. Z value generated in the previous section.

        Please note that in this function, we only take z value in ONLY ONE CHANNEL.

        title: A string. Title of the plot.
        """
        logexpTime = np.log(self.exposures*(10**-6))
        fig = plt.figure()
        plt.title(title)
        plt.xlabel('Log exposure')
        plt.ylabel('Pixel intensity value')
        LEx = np.expand_dims(self.le, axis=1)
        LEx = np.repeat(LEx, logexpTime.shape[0], axis=1)
        logx = np.expand_dims(logexpTime, axis=1)
        logx = np.swapaxes(logx, 0, 1)
        logx = np.repeat(logx, self.le.shape[0], axis=0)
        x = logx + LEx
        plt.plot(x, self.raw_samples, 'ro', alpha=0.5)
        plt.plot(self.g, np.linspace(0, 255, 256))
        
        if not os.path.exists('CapturedImages/'):
            os.mkdir('CapturedImages/')
        if not os.path.exists('CapturedImages/sequenceImages/'):
            os.mkdir('CapturedImages/sequenceImages/')
        if not os.path.exists('CapturedImages/sequenceImages/undistortRadioCalib/'):
            os.mkdir('CapturedImages/sequenceImages/undistortRadioCalib/')
        if not os.path.exists('CapturedImages/sequenceImages/undistortRadioCalib/radioCalibResults/'):
            os.mkdir('CapturedImages/sequenceImages/undistortRadioCalib/radioCalibResults/')
        fig.savefig('CapturedImages/sequenceImages/undistortRadioCalib/radioCalibResults/Camera response.png')
        print('Camera Response plot successful! Plot viewable at CapturedImages/sequenceImages/undistortRadioCalib/radioCalibResults')
        # plt.show()

    def get_camera_response(self, smoothness):
        """
        Some explanation for solving g:
        Given a set of pixel values observed for several pixels in several
        images with different exposure times, this function returns the
        imaging system's response function g as well as the log film irradiance
        values for the observed pixels.
        Assumes:
        Zmin = 0
        Zmax = 255
        Arguments:
        self.raw_sample - Z(i, j)  is the pixel values of pixel location number i in image j
        self.exposure B(j)    is the log delta t, or log shutter speed, for image j
        l       is the lamda, the constant that determines the amount of smoothness
        w(z)    is the weighting function value for pixel value z
        Returns:
        g(z)    is the log exposure corresponding to pixel value z
        lE(i)   is the log film irradiance at pixel location i
        """
        # Load raw data
        self.load_raw_data()
        Z = self.raw_samples.astype(np.int)
        # Convert exposure to log exposure
        B = np.log(self.exposures*(10**-6))
        # Next is to calculate g function #
        n = 256
        # Create weighting function - hat like
        """
        self.w = np.ones([256, 1])
        for i in range(128):
            self.w[i] = i + 1
        for i in range(128, 255):
            self.w[i] = 256 - i
        """
        self.w = np.ones((n, 1)) / n
        m = Z.shape[0]
        p = Z.shape[1]
        A = np.zeros((m * p + n + 1, n + m))
        b = np.zeros((A.shape[0], 1))
        k = 0
        # Data fitting equations
        for i in range(m):
            for j in range(p):
                wij = self.w[Z[i, j]]
                A[k, Z[i, j]] = wij
                A[k, n + i] = -wij
                b[k, 0] = wij * B[j]
                k += 1
        # Fix the curve by setting its middle value to 0
        A[k, 128] = 1
        k = k + 1
        # Include smoothness equations
        for i in range(n - 2):
            A[k, i] = smoothness * self.w[i + 1]
            A[k, i + 1] = -2 * smoothness * self.w[i + 1]
            A[k, i + 2] = smoothness * self.w[i + 1]
            k = k + 1
        # Solve the system using SVD
        x = np.linalg.lstsq(A, b, rcond=None)
        x = x[0]
        self.g = x[0:n]
        lE = x[n:x.shape[0]]
        self.le = lE.squeeze()
        # Save g function, exposures, etc for loading
        np.savez('CalibrationNumpyData/radiometric.npz', g_function=self.g, log_exposures=self.le[::10],
                 w_function=self.w, exposures=self.exposures, samples=self.raw_samples[::10, :])
        return self.g, self.le

    def get_HDR_image(self, images=None, exposures=None):
        # If images is None, take radiometric calibration images
        if images is None:
            if self.raw_data is None:
                self.load_raw_data()
                images = self.raw_data
            else:
                images = self.raw_data
        # If g function is None, load calibration
        if self.g is None:
            self.load_calibration_data()
        # Override exposure values
        if exposures is not None:
            self.exposures = exposures
        # Compute log exposure image
        # Initialize flatten
        size = (int(self.height * 1), int(self.width * 1))
        EE = np.zeros([size[0] * size[1], 1])
        sumw = np.zeros([size[0] * size[1], 1], int)
        # Convert exposure from microseconds to seconds
        exp_sec = self.exposures * (10 ** -6)
        num_exp = self.exposures.shape[0]
        for i in range(num_exp):
            t = images[i].flatten()
            EE = EE + self.w[t] * (self.g[t] - np.log(exp_sec[i]))
            #EE = EE + (self.g[t] - np.log(exp_sec[i]))
            sumw = sumw + self.w[t]
        # Reshape
        lE = np.reshape(EE / sumw, size)
        #lE = np.reshape(EE / num_exp, size)
        # Take exponent to get exposure for each pixel
        exposure_image = np.exp(lE)
        return exposure_image

    def calibrate_image(self, exposure, path):
        # UNUSED
        # Create list of calibrated images
        images = []
        # Exposure in microseconds -> convert to seconds
        exp = exposure * (10 ** -6)
        g = np.exp(self.g)
        # Load images
        imgFileList = self.readFileList(path)
        # Idx
        k = 0
        # Iterate over images to be calibrated
        for i in imgFileList:
            # Read image and convert to grayscale if necessary
            img = cv2.imread(i)
            if img.shape[2] == 3:
                gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
            else:
                gray = img
            # Applying the Debevec Algorithm
            # Eq. 5 Debevec et. al.
            calibrated_image = g[gray]
            #calibrated_image = np.exp(calibrated_image - np.log(exp))
            calibrated_image = calibrated_image - np.log(exp)
            #calibrated_image *= 255.0 / calibrated_image.max()
            images.append(calibrated_image)
            k += 1
        # Normalize by last captured image, which represents the object lit by a constant (255) illumination pattern
        illuminated_radiance = images.pop()
        for r in range(len(images)):
            n_img = images[r]/illuminated_radiance
            #n_img = images[r]
            # Gamma correction
            #n_img = (n_img - np.min(n_img)) / (np.max(n_img) - np.min(n_img))
            n_img = self.apply_gamma_curve(n_img, gamma=0.4)
            cv2.imwrite(path + '/RadianceMaps/capture' + str(r) + '.PNG', n_img*255)
            np.save(path + '/RadianceMaps/capture_' + str(r) + '.npy', n_img)
            images[r] = n_img
        return images, g

    @staticmethod
    def scaleBrightness(E):
        # Unused
        """
        Brightness scaling function, which will scale the values on the radiance map to between 0 and 1

        Args:
            E: An m*n*3 array. m*n is the size of your radiance map, and 3 represents R, G and B channel. It is your plotted Radiance map (don't forget to use np.exp function to get it back from logorithm of radiance!)
        Returns:
            ENomrMap: An m*n*3 array. Normalized radiance map, whose value should between 0 and 1
        """
        res = np.zeros(E.shape)
        for c in range(E.shape[2]):
            res[:, :, c] = (E[:, :, c] - np.min(E[:, :, c])) / (np.max(E[:, :, c]) - np.min(E[:, :, c]))
        return res

    @staticmethod
    def apply_gamma_curve(E, gamma=0.4):
        # Unused
        """
        apply gamma to the curve through raising E to the gamma.

        Args:
            E: An m*n*3 array. m*n is the size of your radiance map, and 3 represents R, G and B channel. It is your plotted Radiance map (don't forget to use np.exp function to get it back from logorithm of radiance!)
            gamma: a float value that is representative of the power to raise all E to.
        Returns:
            E_gamma: E modified by raising it to gamma.
        """
        return E ** gamma

    @staticmethod
    def readFileList(imgFolder, ImgPattern="*.PNG"):
        imgFileList = glob.glob(os.path.join(imgFolder, ImgPattern))
        imgFileList.sort()
        print(imgFileList)
        return imgFileList




import numpy as np
import cv2
from Camera import Camera
import time
from abc import ABC
from selenium import webdriver
from selenium.webdriver.common.keys import Keys
import subprocess
import os
import signal




class IpCam(Camera, ABC):

    def __init__(self, ipCamPath='/Users/Merlin/Documents/GitHub/IPcam/',
                 excChromePath='/Users/Merlin/Downloads/chromedriver', url='http://localhost:1234/', roomid='111',
                 exposure=0.1, white_balance=0, auto_focus=True, grayscale=True):
        # Setting and initializing the external camera
        resolution = (4032, 3024)
        fps = 1
        # Init base class
        super().__init__(exposure, white_balance, auto_focus, fps, resolution, grayscale)
        self.ipCamPath = ipCamPath
        # Start server running bash script
        ipCamPathScript = ipCamPath + 'ios-ip-cam.sh ' + ipCamPath
        #rc = subprocess.call(ipCamPathScript, shell=True)
        self.pro = subprocess.Popen(ipCamPathScript, stdout=subprocess.PIPE, shell=True, preexec_fn=os.setsid)
        # Grab current frame
        self.driver = webdriver.Chrome(executable_path=excChromePath)
        # /Users/Merlin/Downloads/chromedriver
        #self.driver = webdriver.Chrome()
        # Go to your page url
        #url = 'http://google.com/'
        self.driver.get(url)
        # Enter room id
        username = self.driver.find_element_by_css_selector('input.MuiInputBase-input.MuiInput-input')
        username.send_keys(roomid)
        # Enter room
        enter = self.driver.find_element_by_css_selector('button.MuiButtonBase-root.MuiButton-root.MuiButton-outlined.makeStyles-button-3')
        enter.click()
        connect = self.driver.find_element_by_css_selector('button.MuiButtonBase-root.MuiButton-root.MuiButton-outlined.Video-button-184')
        connect.click()
        time.sleep(10)
        self.connected = True
        self.hdr_exposures = None

    def getFPS(self):
        # Returns the frame rate
        return self.fps

    def setFPS(self, fps):
        # Sets the on frame rate
        self.fps = fps

    def setAutoGain(self):
        raise NotImplementedError

    def getGain(self):
        raise NotImplementedError

    def setGain(self):
        raise NotImplementedError

    def getResolution(self):
        # Returns a tuple resolution (width, height)
        return self.resolution

    def setResolution(self, resolution):
        # Sets a new resolution
        # May only support certain resolutions
        self.resolution = resolution

    def getImage(self, name='test', saveImage=True, saveNumpy=True, calibration=False, timeout=5000):
        # Get button you are going to click by its id ( also you could us find_element_by_css_selector to get element by css selector)
        if self.connected:
            disconnect = self.driver.find_element_by_css_selector('button.MuiButtonBase-root.MuiButton-root.MuiButton-outlined.Video-button-184')
            disconnect.click()
            self.connected = False
        #take_photo = self.driver.find_element_by_css_selector('span.MuiButton-endIcon.MuiButton-iconSizeMedium')
        #take_photo.click()
        cap_num = name[-1]
        img_path = self.ipCamPath + 'ip-cam/results/' + cap_num + '.png'
        while not os.path.isfile(img_path):
            time.sleep(1)
        frame = cv2.imread(img_path)
        print(name, " :", frame.shape)
        if saveImage:
            if calibration:
                cv2.imwrite('CalibrationImages/' + name + '.PNG', frame)
            else:
                cv2.imwrite('CapturedImages/' + name + '.PNG', frame)
        if saveNumpy:
            if calibration:
                np.save('CalibrationNumpyData/' + name, frame)
            else:
                np.save('CapturedNumpyData/' + name, frame)
        return frame

    def setAutoExposure(self):
        raise NotImplementedError

    def setExposure(self, exposure):
        raise NotImplementedError

    def getExposure(self):
        raise NotImplementedError

    def viewCameraStream(self):
        raise NotImplementedError

    def quit_and_close(self):
        self.pro.kill()
        #os.killpg(os.getpgid(self.pro.pid), signal.SIGTERM)

    def quit_and_open(self):
        raise NotImplementedError

    def getStatus(self):
        raise NotImplementedError



from abc import ABC
from GhostScan.CalibrationSession import CalibrationSession


class GeometricCalibSession(CalibrationSession, ABC):
    def __init__(self, camera, projector, geometric_calib, path='CalibrationImages'):
        # Set camera
        self.camera = camera
        # Set projector
        self.projector = projector
        self.path = path
        self.geometric_calib = geometric_calib

    def capture(self):
        # For the extrinsic/geometric calibration you are required to print 8 Aruco markers
        # Checkerboard: CalibrationImages/6x6_40mm.pdf
        # Glue or mount them on the edges of you mirror as well as halfway between the edges
        # The Aruco markers will determine the position of the object relative to the camera
        # To determine the position of the screen, we display a checkerboard pattern on the screen
        # Make sure that the screen is visible on the mirror from the position of the camera
        self.projector.displayCalibrationPattern(self.camera)
        # The image is by default saved in CalibrationImages/Geometric

    def calibrate(self):
        # TODO: Include all calls necessary for calibration
        self.geometric_calib.calibrate()
        return self.camera



from abc import abstractmethod


class Camera:
    def __init__(self, exposure, white_balance, auto_focus, fps, resolution, grayscale):
        # Exposure passed as float in seconds
        self.exposure = exposure
        # White balanced passed as a float
        self.white_balance = white_balance
        # Auto_focus passed as boolean
        self.auto_focus = auto_focus
        # FPS in float
        self.fps = fps
        # Resolution as tuple (Width, Height)
        self.resolution = resolution
        # Grayscale in boolean
        self.grayscale = grayscale
        # Capture object may be in cv2.capture, pypylon, PySpin etc.
        self.capture = None
        # Calibration object
        self.calibration = None

    @abstractmethod
    def getImage(self):
        # To override: Capture image, return frame and save in corresponding folder in specified file format
        pass

    @abstractmethod
    def setExposure(self, exposure):
        # To override
        pass

    @abstractmethod
    def getExposure(self):
        # To override
        pass

    @abstractmethod
    def getFPS(self):
        # To override
        pass

    @abstractmethod
    def setFPS(self):
        # To override
        pass

    @abstractmethod
    def setAutoGain(self):
        # To override
        pass

    @abstractmethod
    def getGain(self):
        # To override
        pass

    @abstractmethod
    def setGain(self):
        # To override
        pass

    @abstractmethod
    def getResolution(self):
        # To override
        pass

    @abstractmethod
    def setResolution(self):
        # To override
        pass

    @abstractmethod
    def viewCameraStream(self):
        # To override
        pass

    @abstractmethod
    def quit_and_close(self):
        # To override
        pass

    @abstractmethod
    def quit_and_open(self):
        # To override
        pass

    @abstractmethod
    def getStatus(self):
        # To override
        pass

    @abstractmethod
    def setCalibration(self, calibration):
        # Set the calibration object
        self.calibration = calibration








import numpy as np
import cv2
from GhostScan.Camera import Camera
import time
import matplotlib.pyplot as plt
import sys
from abc import ABC
# Just for Jupyter
from io import BytesIO, StringIO
from IPython.display import clear_output, Image, display, update_display
from bokeh.plotting import figure
from bokeh.io import output_notebook, show, push_notebook
import PIL
import os


class Internal(Camera, ABC):

    def __init__(self, exposure=0.1, white_balance=0, auto_focus=True, grayscale=True, internal_camera_port=1):
        # Setting and initializing the external camera
        self.port = internal_camera_port
        self.cap = cv2.VideoCapture(self.port)
        if self.cap.isOpened():
            # Wait for camera to initialize, otherwise black frames may occur
            time.sleep(0.5)
        if self.cap is None or not self.cap.isOpened():
            print("Warning: No internal webcam found.")
        # Get framerate and resolution of camera
        fps = self.getFPS()
        resolution = self.getResolution()
        self.grayscale = grayscale
        # Init base class
        super().__init__(exposure, white_balance, auto_focus, fps, resolution, grayscale)
        self.hdr_exposures = None

    def getFPS(self):
        # Returns the frame rate
        fps = self.cap.get(cv2.CAP_PROP_FPS)
        return fps

    def setFPS(self, fps):
        # Sets the on frame rate
        self.cap.set(cv2.CAP_PROP_FPS, self.fps)
        self.fps = fps

    def setAutoGain(self):
        raise NotImplementedError

    def getGain(self):
        raise NotImplementedError

    def setGain(self):
        raise NotImplementedError

    def getResolution(self):
        # Returns a tuple resolution (width, height)
        width = int(self.cap.get(cv2.CAP_PROP_FRAME_WIDTH))
        height = int(self.cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
        return width, height

    def setResolution(self, resolution):
        # Sets a new resolution
        # May only support certain resolutions
        self.cap.set(cv2.CAP_PROP_FRAME_WIDTH, self.resolution[1])
        self.cap.set(cv2.CAP_PROP_FRAME_HEIGHT, self.resolution[0])
        self.resolution = resolution

    def getImage(self, name="", saveImage=True, saveNumpy=True, calibration=False, timeout=5000):
        
        # Grab current frame
        ret, frame = self.cap.read()
        # Save images
        if saveImage:
            if calibration:
                cv2.imwrite('CalibrationImages/' + name + '.PNG', cv2.cvtColor(frame, cv2.COLOR_BGR2RGB))
            else:
                cv2.imwrite('CapturedImages/' + name + '.PNG', cv2.cvtColor(frame, cv2.COLOR_BGR2RGB))
        if saveNumpy:
            if calibration:
                np.save('CalibrationNumpyData/' + name, cv2.cvtColor(frame, cv2.COLOR_BGR2RGB))
            else:
                np.save('CapturedNumpyData/' + name, cv2.cvtColor(frame, cv2.COLOR_BGR2RGB))
        return frame


    def setAutoExposure(self):
        # Turn on auto exposure
        self.cap.set(cv2.CAP_PROP_AUTO_EXPOSURE, 0.75)

    def setExposure(self, exposure):
        # Not settable on most built in devices
        self.cap.set(cv2.CAP_PROP_IOS_DEVICE_EXPOSURE, exposure)
        self.exposure = exposure

    def getExposure(self):
        # Not gettable on most built in devices
        return self.cap.get(cv2.CAP_PROP_IOS_DEVICE_EXPOSURE)

    def viewCameraStream(self):
        # Live view
        while True:
            img = self.getImage(saveImage=False, saveNumpy=False)
            cv2.imshow('Internal Webcam Stream', img)
            c = cv2.waitKey(1)
            if c != -1:
                # When everything done, release the capture
                cv2.destroyAllWindows()
                self.quit_and_open()
                break

    def viewCameraStreamJupyter(self):
        # Live view in a Jupyter Notebook
        try:
            while True:
                start = self.getImage(saveImage=False, saveNumpy=False)
                start = cv2.cvtColor(start,cv2.COLOR_BGR2RGB)
                g = BytesIO()
                PIL.Image.fromarray(start).save(g, 'jpeg')
                obj = Image(data=g.getvalue())
                dis = display(obj, display_id=True)
                clear_output(wait=True)
        except KeyboardInterrupt: 
            self.quit_and_open()

    def viewCameraStreamJupyterBokeh(self):
        try:
            # Live view in a Jupyter Notebook
            # Flickering occurs
            output_notebook()
            frame = self.getImage(saveImage=False, saveNumpy=False)
            frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGBA)  # because Bokeh expects a RGBA image
            frame = cv2.flip(frame, -1)  # because Bokeh flips vertically
            width = frame.shape[1]
            height = frame.shape[0]
            p = figure(x_range=(0, width), y_range=(0, height), width=width, height=height)
            myImage = p.image_rgba(image=[frame], x=0, y=0, dw=width, dh=height)
            show(p, notebook_handle=True)
            while True:
                frame = self.getImage(saveImage=False, saveNumpy=False)
                frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGBA)
                frame = cv2.flip(frame, -1)
                myImage.data_source.data['image'] = [frame]
                push_notebook()
                time.sleep(2)
        except KeyboardInterrupt:
            self.quit_and_open()

    def quit_and_close(self):
        # Close camera
        self.cap.release()

    def quit_and_open(self):
        # Close camera
        self.cap.release()
        # Create new capture
        self.cap = cv2.VideoCapture(self.port)

    def getStatus(self):
        if self.cap is None or not self.cap.isOpened():
            print('Warning: unable to open video source port: ', self.port)


class External(Camera):

    def __init__(self, exposure=0.01, white_balance=0, auto_focus=False, grayscale=True, external_camera_port=0):
        # Setting and initializing the external camera
        self.port = external_camera_port
        self.cap = cv2.VideoCapture(self.port)
        if self.cap.isOpened():
            # Wait for camera to initialize, otherwise black frames may occur
            time.sleep(0.5)
        if self.cap is None or not self.cap.isOpened():
            print('Warning: unable to open external video source port: ', self.port)
        # Get framerate and resolution of camera
        fps = self.getFPS()
        resolution = self.getResolution()
        self.grayscale = grayscale
        # Init base class
        super().__init__(exposure, white_balance, auto_focus, fps, resolution, grayscale)
        self.hdr_exposures = None

    def setAutoExposure(self):
        # Turn on auto exposure
        self.cap.set(cv2.CAP_PROP_AUTO_EXPOSURE, 0.75)

    def getFPS(self):
        # Returns the frame rate
        fps = self.cap.get(cv2.CAP_PROP_FPS)
        return fps

    def setFPS(self, fps):
        # Not settable on Mac
        self.cap.set(cv2.CAP_PROP_FPS, self.fps)
        self.fps = fps

    def setAutoGain(self):
        raise NotImplementedError

    def getGain(self):
        raise NotImplementedError

    def setGain(self):
        raise NotImplementedError

    def getResolution(self):
        # Returns a tuple resolution (width, height)
        width = self.cap.get(cv2.CAP_PROP_FRAME_WIDTH)
        height = self.cap.get(cv2.CAP_PROP_FRAME_HEIGHT)
        return width, height

    def setResolution(self, resolution):
        # Sets new tuple of resolution
        self.cap.set(cv2.CAP_PROP_FRAME_WIDTH, self.resolution[1])
        self.cap.set(cv2.CAP_PROP_FRAME_HEIGHT, self.resolution[0])
        self.resolution = resolution

    def setExposure(self, exposure):
        # Turn off auto exposure
        self.cap.set(cv2.CAP_PROP_AUTO_EXPOSURE, 0.25)
        # Set exposure
        self.cap.set(cv2.CAP_PROP_IOS_DEVICE_EXPOSURE, exposure)
        self.exposure = exposure

    def getExposure(self):
        # Return exposure value
        return self.cap.get(cv2.CAP_PROP_IOS_DEVICE_EXPOSURE)

    def getImage(self, name='test', saveImage=True, saveNumpy=True, calibration=False, timeout=5000):
        # Grab current frame
        ret, frame = self.cap.read()
        # Save images
        if saveImage:
            if calibration:
                cv2.imwrite('CalibrationImages/' + name + '.PNG', cv2.cvtColor(frame, cv2.COLOR_BGR2RGB))
            else:
                cv2.imwrite('CapturedImages/' + name + '.PNG', cv2.cvtColor(frame, cv2.COLOR_BGR2RGB))
        if saveNumpy:
            if calibration:
                np.save('CalibrationNumpyData/' + name, cv2.cvtColor(frame, cv2.COLOR_BGR2RGB))
            else:
                np.save('CapturedNumpyData/' + name, cv2.cvtColor(frame, cv2.COLOR_BGR2RGB))
        return frame

    def viewCameraStream(self):
        while True:
            img = self.getImage(saveImage=False, saveNumpy=False)
            cv2.imshow('Internal Webcam Stream', img)
            c = cv2.waitKey(1)
            if c != -1:
                # When everything done, release the capture
                self.quit_and_open()
                cv2.destroyAllWindows()
                break

    def viewCameraStreamJupyter(self):
        # Live view in a Jupyter Notebook
        try:
            while True:
                start = self.getImage(saveImage=False, saveNumpy=False)
                start = cv2.cvtColor(start,cv2.COLOR_BGR2RGB)
                g = BytesIO()
                PIL.Image.fromarray(start).save(g, 'jpeg')
                obj = Image(data=g.getvalue())
                dis = display(obj, display_id=True)
                clear_output(wait=True)
        except KeyboardInterrupt: 
            self.quit_and_open()

    def quit_and_close(self):
        # Close camera
        self.cap.release()

    def quit_and_open(self):
        # Close camera
        self.cap.release()
        # Create new capture
        self.cap = cv2.VideoCapture(self.port)

    def getStatus(self):
        if self.cap is None or not self.cap.isOpened():
            print('Warning: unable to open external video source port: ', self.port)



import sys
import sys
import glob
import os
import math
import numpy as np
import cv2
from cv2 import aruco
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D


class GeometricCalibration:
    def __init__(self, cam, projector, intrinsicCalibFile='CalibrationNumpyData/intrinsic_calibration.npz',
                 imgFile='CalibrationImages/Geometric/geo.PNG', checker_pattern=(23, 7), checker_size=0.01,
                 checker_pixels=60,
                 width_mirror=0.3556, height_mirror=0.254, marker_offset=0.005, dict_aruco=aruco.DICT_6X6_250,
                 aruco_size=0.04, imgPattern='*.PNG'):
        # For the extrinsic/geometric calibration you are required to print 8 Aruco markers
        # Checkerboard: CalibrationImages/6x6_40mm.pdf
        # Glue or mount them on the edges of you mirror as well as halfway between the edges
        # The Aruco markers will determine the position of the object relative to the camera
        # To determine the position of the screen, we display a checkerboard pattern on the screen
        # Make sure that the screen is visible on the mirror from the position of the camera
        # Camera:
        self.cam = cam
        # Projector:
        self.proj = projector
        # Checkerboard:
        # Tuple of checkers in width and height
        self.checker_pattern = checker_pattern
        # Checker file:
        self.checker_file = 'CalibrationNumpyData/8_24_checker.npz'
        # Display size of checker
        self.checker_size = checker_size
        # Pixel size
        self.checker_pixels = checker_pixels
        # Per pixel Size
        self.checker_pixel_size = checker_size / checker_pixels
        # Display dimensions:
        self.display_width, self.display_height = self.proj.getResolution()
        # Aruco marker dimension:
        self.aruco_size = aruco_size
        self.aruco_dict = aruco.Dictionary_get(dict_aruco)
        # Mirror/Object dimensions: height and length of mirror
        # Subtract the white marker offset on a marker patch
        self.board_width = width_mirror
        self.board_height = height_mirror
        self.half_width = (width_mirror / 2) - marker_offset
        self.half_height = (height_mirror / 2) - marker_offset
        # TODO: paths
        # File pattern of images
        self.imgPattern = imgPattern
        self.imgFile = imgFile
        self.intrinsicFile = intrinsicCalibFile
        # self.imgUndistortFile = imgUndistortFile

    def readFileList(self, imgFolder):
        # Yunhao Li
        imgFileList = glob.glob(os.path.join(imgFolder, self.imgPattern))
        imgFileList.sort()
        return imgFileList

    def detectChecker(self, img, debug=True):
        # Yunhao
        if len(img.shape) == 3:
            gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
        elif len(img.shape) == 2:
            gray = img

        ret, corners = cv2.findChessboardCorners(gray, self.checker_pattern,
                                                 cv2.CALIB_CB_ADAPTIVE_THRESH + cv2.CALIB_CB_EXHAUSTIVE)
        corners_refine = corners

        if ret:
            criteria = (cv2.TERM_CRITERIA_EPS + cv2.TERM_CRITERIA_MAX_ITER, 30, 0.001)
            corners_refine = cv2.cornerSubPix(gray, corners, (3, 3), (-1, -1), criteria)

            if debug:
                cv2.drawChessboardCorners(img, self.checker_size, corners_refine, ret)
                cv2.namedWindow('Checker', cv2.WINDOW_NORMAL)
                cv2.imshow('Checker', img)
                cv2.waitKey(0)  # any key
                cv2.destroyWindow('Checker')

        return ret, corners_refine

    @staticmethod
    def readCheckerObjPoint(fname):
        # Yunhao Li
        data = np.load(fname)
        objp = data["objp"]
        return objp

    def arucoBoard(self):
        # Yunhao Li
        # Aruco marker length, board height, board width
        m = self.aruco_size
        h = self.board_height
        w = self.board_width
        # create objPoints for calibration target
        h0 = (0 - h / 2)
        hm = (m - h / 2)
        h1 = (((h - m) / 2) - h / 2)
        h2 = (((h + m) / 2) - h / 2)
        h3 = ((h - m) - h / 2)
        h4 = (h - h / 2)
        w0 = (0 - w / 2)
        wm = (m - w / 2)
        w1 = (((w - m) / 2) - w / 2)
        w2 = (((w + m) / 2) - w / 2)
        w3 = ((w - m) - w / 2)
        w4 = (w - w / 2)

        objPoints = []
        objPoints.append(np.array([[w0, h0, 0], [wm, h0, 0], [wm, hm, 0], [w0, hm, 0]], dtype=np.float32))  # 0
        objPoints.append(np.array([[w0, h1, 0], [wm, h1, 0], [wm, h2, 0], [w0, h2, 0]], dtype=np.float32))  # 1
        objPoints.append(np.array([[w0, h3, 0], [wm, h3, 0], [wm, h4, 0], [w0, h4, 0]], dtype=np.float32))  # 2
        objPoints.append(np.array([[w1, h3, 0], [w2, h3, 0], [w2, h4, 0], [w1, h4, 0]], dtype=np.float32))  # 3
        objPoints.append(np.array([[w3, h3, 0], [w4, h3, 0], [w4, h4, 0], [w3, h4, 0]], dtype=np.float32))  # 4
        objPoints.append(np.array([[w3, h1, 0], [w4, h1, 0], [w4, h2, 0], [w3, h2, 0]], dtype=np.float32))  # 5
        objPoints.append(np.array([[w3, h0, 0], [w4, h0, 0], [w4, hm, 0], [w3, hm, 0]], dtype=np.float32))  # 6
        objPoints.append(np.array([[w1, h0, 0], [w2, h0, 0], [w2, hm, 0], [w1, hm, 0]], dtype=np.float32))  # 7

        ids = np.linspace(0, 7, 8).astype(np.int32)[:, None]
        arucoCornerBoard = aruco.Board_create(objPoints, self.aruco_dict, ids)

        return arucoCornerBoard, objPoints

    def detectAruco(self, img, debug=True):
        # Yunhao Li
        if len(img.shape) == 3:
            gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
        elif len(img.shape) == 2:
            gray = img

        parameters = aruco.DetectorParameters_create()
        # corners, ids, rejectedImgPoints = aruco.detectMarkers(img, aruco_dict, parameters=parameters)
        corners, ids, rejectedImgPoints = aruco.detectMarkers(gray, self.aruco_dict, parameters=parameters)

        if debug:
            frame_markers = aruco.drawDetectedMarkers(img.copy(), corners, ids)
            cv2.namedWindow('Aruco', cv2.WINDOW_NORMAL)
            cv2.imshow('Aruco', frame_markers)
            cv2.waitKey(0)  # any key
            cv2.destroyWindow('Aruco')

        return corners, ids

    def postEst(self, corners, ids, camMat, distCoeffs):
        # Yunhao Li
        arucoCornerBoard, _ = self.arucoBoard()
        retval, rvec, tvec = aruco.estimatePoseBoard(corners, ids, arucoCornerBoard, camMat, distCoeffs, None, None)
        return rvec, tvec

    def reProjAruco(self, img, camMat, distCoeffs, rvec, tvec, cornersAruco):
        # Yunhao Li
        print("reProjAruco")
        _, objPoints = self.arucoBoard()  # yunhao

        ids = np.linspace(0, 7, 8).astype(np.int32)[:, None]
        corners_reproj = []
        for i in range(len(objPoints)):
            imgPoints, _ = cv2.projectPoints(np.array(objPoints[i]), rvec, tvec, camMat, distCoeffs)
            corners_reproj.append(imgPoints)

        frame_markers = aruco.drawDetectedMarkers(img.copy(), corners_reproj, ids)
        cv2.imwrite("./reproejct_markers.png", frame_markers)
        cv2.namedWindow('Reproject', cv2.WINDOW_NORMAL)
        cv2.imshow('Reproject', frame_markers)
        cv2.waitKey(0)  # any key
        cv2.destroyWindow('Reproject')

    @staticmethod
    def householderTransform(n, d):
        # Yunhao Li
        I3 = np.identity(3, dtype=np.float32)
        e = np.array([0, 0, 1])
        p1 = I3 - 2 * np.outer(n, n)
        p2 = I3 - 2 * np.outer(e, e)
        p3 = 2 * d * n
        return p1, p2, p3

    @staticmethod
    def invTransformation(R, t):
        # Yunhao Li
        Rinv = R.T
        Tinv = -(Rinv @ t)
        return Rinv, Tinv

    def calib(self, imgPath, cameraCalibPath):
        # Yunhao Li
        # imgFileList = readFileList(imgFolder)

        data = np.load(cameraCalibPath)
        camMtx = data["mtx"]
        dist = data["dist"]

        objP_pixel = np.ceil(self.readCheckerObjPoint(self.checker_file))
        objP_pixel[:, 2] = 0
        objP = np.array(objP_pixel)
        for i in range(self.checker_pattern[1]):
            for j in range(math.floor(self.checker_pattern[0] / 2)):
                tmp = objP[self.checker_pattern[0] * i + j, 0]
                objP[self.checker_pattern[0] * i + j, 0] = objP[
                    self.checker_pattern[0] * i + self.checker_pattern[0] - j - 1, 0]
                objP[self.checker_pattern[0] * i + self.checker_pattern[0] - j - 1, 0] = tmp
        objP[:, 0] -= (self.display_width / 2 - 1)
        objP[:, 1] -= (self.display_height / 2 - 1)

        objP *= self.checker_size

        rtA = []
        rB = []
        tB = []
        rC2Ss = []
        tC2Ss = []

        # define valid image
        validImg = -1
        # for i in trange(len(imgFileList), desc="Images"):
        for i in range(1):

            img = cv2.imread(imgPath, cv2.IMREAD_UNCHANGED)

            # Yunhao
            # img = (img/65535*255).astype(np.uint8)

            # Aruco marker for Mirror position
            cornersAruco, ids = self.detectAruco(img, debug=False)
            if cornersAruco is None and ids is None and len(cornersAruco) <= 3:
                continue

            # Checker for Display
            ret, cornersChecker = self.detectChecker(img, debug=False)
            if not ret:
                print("no Checker!!!")
                continue

            # for a valid image, aruco and checker must be both detected
            validImg += 1

            # Calibrate Mirror Pose with Aruco

            rvecMirror, tvecMirror = self.postEst(cornersAruco, ids, camMtx, dist)
            img_axis = aruco.drawAxis(img, camMtx, dist, rvecMirror, tvecMirror, self.aruco_size)
            cv2.namedWindow('Img_axis', cv2.WINDOW_NORMAL)
            cv2.imshow('Img_axis', img_axis)
            cv2.waitKey(0)  # any key
            cv2.destroyWindow('Img_axis')

            ## Reproejct Camera Extrinsic
            self.reProjAruco(img, camMtx, dist, rvecMirror, tvecMirror, cornersAruco)

            rMatMirror, _ = cv2.Rodrigues(rvecMirror)  # rotation vector to rotation matrix
            normalMirror = rMatMirror[:, 2]

            rC2W, tC2W = self.invTransformation(rMatMirror, tvecMirror)
            dW2C = abs(np.dot(normalMirror, tvecMirror))

            # Householder transformation
            p1, p2, p3 = self.householderTransform(normalMirror, dW2C)

            # Calibrate virtual to Camera with Checker
            rpe, rvecVirtual, tvecVirtual = cv2.solvePnP(objP, cornersChecker, camMtx, dist,
                                                         flags=cv2.SOLVEPNP_IPPE)  # cv2.SOLVEPNP_IPPE for 4 point solution #cv2.SOLVEPNP_ITERATIVE
            # iterationsCount=200, reprojectionError=8.0,

            rvecVirtual, tvecVirtual = cv2.solvePnPRefineLM(objP, cornersChecker, camMtx, dist, rvecVirtual,
                                                            tvecVirtual)

            proj, jac = cv2.projectPoints(objP, rvecVirtual, tvecVirtual, camMtx, dist)
            img_rep = img

            cv2.drawChessboardCorners(img_rep, self.checker_pattern, proj, True)
            width = 960
            height = int(img_rep.shape[0] * 960 / img_rep.shape[1])
            smallimg = cv2.resize(img_rep, (width, height))
            cv2.imshow("img_rep", smallimg)
            cv2.waitKey(0)  # any key
            cv2.destroyWindow("img_rep")

            rMatVirtual, _ = cv2.Rodrigues(rvecVirtual)  # rotation vector to rotation matrix

            print(tvecVirtual)
            if validImg == 0:
                rtA = p1
                rB = np.matmul(rMatVirtual, p2)
                tB = np.squeeze(tvecVirtual) + p3
            else:
                rtA = np.concatenate((rtA, p1))
                rB = np.concatenate((rB, np.matmul(rMatVirtual, p2)))
                tB = np.concatenate((tB, np.squeeze(tvecVirtual) + p3))

            rS2C = p1 @ rMatVirtual
            tS2C = p1 @ np.squeeze(tvecVirtual) + p3

            rC2S, tC2S = self.invTransformation(rS2C, tS2C)
            print("rC2S:", rC2S)
            print("tC2S:", tC2S)
            rC2Ss.append(rC2S)
            tC2Ss.append(tC2S)

        # rC2Ss = np.array(rC2Ss)
        # tC2Ss = np.array(tC2Ss)
        # fout = os.path.join(imgFolder, "Cam2Screen.npz")
        # np.savez(fout, rC2S=rC2Ss, tC2S=tC2Ss)
        return rC2Ss, tC2Ss

    def calibrate(self, imgUndistortFile):
        # Yunhao Li
        # Load intrinsic matrix npz file
        intrinsic_mat = np.load(self.intrinsicFile)

        # Checkerboard:
        CHECKSQUARE_SIZE = self.checker_size
        CHECKSQUARE_PIXELS = self.checker_pixels
        # Display dimensions:
        SCREEN_WIDTH = self.display_width
        SCREEN_HEIGHT = self.display_height

        # half_length = 0.1778 - 0.005
        # half_height = 0.127 - 0.005

        rc2s, tc2s = self.calib(self.imgFile, self.intrinsicFile)
        tc2s = tc2s[0]
        print("new rc2s is ", rc2s)
        print("new tc2s is ", tc2s)
        # rcamera_dis = cam2screen.f.rC2S
        # tcamera_dis = cam2screen.f.tC2S
        rcamera_dis = rc2s
        tcamera_dis = tc2s
        cameraMatrix = intrinsic_mat.f.mtx
        disCoeffs = intrinsic_mat.f.dist
        ret = intrinsic_mat.f.ret
        # additional section: Make sure that the shape of rC2S and tC2S is correct
        rcamera_dis = rcamera_dis[0]
        tcamera_dis = np.reshape(tcamera_dis, (3,))
        print("rc2s is ", rcamera_dis)
        print("tc2s is ", tcamera_dis)
        print("intrinsic matrix is ", cameraMatrix)
        # print(intrinsic_mat.f.dist)
        allCorners = []
        allIds = []
        decimator = 0
        dictionary = aruco.getPredefinedDictionary(cv2.aruco.DICT_6X6_250)
        # board =
        img = cv2.imread(self.imgFile)
        gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
        cornerRefine = cv2.aruco.CORNER_REFINE_CONTOUR
        [markerCorners, markerIDs, rejectedImgPoints] = cv2.aruco.detectMarkers(gray, dictionary,
                                                                                cameraMatrix=cameraMatrix,
                                                                                distCoeff=disCoeffs)

        # marker_size = 0.04
        board_corner = np.array([np.array(
            [[-self.half_width, -self.half_height, 0], [-self.half_width + 0.04, -self.half_height, 0],
             [-self.half_width + 0.04, -self.half_height + 0.04, 0], [-self.half_width, -self.half_height + 0.04, 0]]),
                                 np.array([[self.half_width - 0.04, -self.half_height, 0],
                                           [self.half_width, -self.half_height, 0],
                                           [self.half_width, -self.half_height + 0.04, 0],
                                           [self.half_width - 0.04, -self.half_height + 0.04, 0]]),
                                 np.array([[self.half_width - 0.04, self.half_height - 0.04, 0],
                                           [self.half_width, self.half_height - 0.04, 0],
                                           [self.half_width, self.half_height, 0],
                                           [self.half_width - 0.04, self.half_height, 0]]),
                                 np.array([[-self.half_width, self.half_height - 0.04, 0],
                                           [-self.half_width + 0.04, self.half_height - 0.04, 0],
                                           [-self.half_width + 0.04, self.half_height, 0],
                                           [-self.half_width, self.half_height, 0]]),
                                 np.array([[-0.02, -self.half_height, 0], [0.02, -self.half_height, 0],
                                           [0.02, -self.half_height + 0.04, 0], [-0.02, -self.half_height + 0.04, 0]]),
                                 np.array([[-self.half_width, -0.02, 0], [-self.half_width + 0.04, -0.02, 0],
                                           [-self.half_width + 0.04, 0.02, 0], [-self.half_width, 0.02, 0]]),
                                 np.array([[self.half_width - 0.04, -0.02, 0], [self.half_width, -0.02, 0],
                                           [self.half_width, 0.02, 0], [self.half_width - 0.04, 0.02, 0]])],
                                dtype=np.float32)

        board_id = np.array([[0], [6], [4], [2], [7], [1], [5]], dtype=np.int32)
        board = cv2.aruco.Board_create(board_corner, dictionary, board_id)

        if len(markerCorners) > 0:
            allCorners.append(markerCorners)
            allIds.append(markerIDs)
            cv2.aruco.drawDetectedMarkers(img, markerCorners, markerIDs, [0, 200, 0])

        # cv2.aruco.drawPlanarBoard(board, img.shape)
        rvecs, tvecs, _objpoints = cv2.aruco.estimatePoseSingleMarkers(markerCorners, 0.04, cameraMatrix, disCoeffs, )
        r_mean = np.mean(rvecs, axis=0)
        t_mean = np.mean(tvecs, axis=0)

        # r_mirror = np.zeros((1,3))
        # t_mirror = np.zeros((1,3))
        r_mirror = None
        t_mirror = None
        ret, r_mirror, t_mirror, = cv2.aruco.estimatePoseBoard(markerCorners, markerIDs, board, cameraMatrix, disCoeffs,
                                                               r_mirror, t_mirror)
        # print("r vec of mirror is ", r_mirror)
        # print("t vec of mirror is ", t_mirror)
        r_mirror_mat = np.zeros((3, 3))
        r_mirror_mat = cv2.Rodrigues(r_mirror, dst=r_mirror_mat, jacobian=None)
        r_mirror_mat = r_mirror_mat[0]
        # print("r mirror mat is ", r_mirror_mat)
        img_axis = aruco.drawAxis(img, cameraMatrix, disCoeffs, r_mirror, t_mirror, 0.04)
        # width = 960
        # height = int(img.shape[0]*960/img.shape[1])
        # smallimg = cv2.resize(img,(width,height))
        cv2.namedWindow('frame', cv2.WINDOW_NORMAL)
        cv2.imshow("frame", img_axis)
        cv2.waitKey(0)  # any key
        cv2.destroyWindow("frame")
        decimator += 1
        imsize = img.shape
        print("size of img is ", imsize)

        # rtcam: camera coordinate system to world coordinate system
        # rtMat: world coordinate system to camera coordinate system
        rtMat = np.hstack((r_mirror_mat, t_mirror))
        rtMat = np.vstack((rtMat, np.array([[0, 0, 0, 1]])))
        print("rtMat is", rtMat)
        rt_cam = np.linalg.inv(rtMat)
        r_cam = rt_cam[0:3, 0:3]
        t_cam = rt_cam[0:3, 3]
        print("rtcam in world coor is ", rt_cam)

        # rtdis: display coordinate system to world coordinate system
        # rtdis_inv: world coordinate system to display coordinate system
        t_camdis = np.reshape(tcamera_dis, (tcamera_dis.shape[0], 1))
        rtMat_camdis = np.hstack((rcamera_dis, t_camdis))
        rtMat_camdis = np.vstack((rtMat_camdis, np.array([0, 0, 0, 1])))
        # rtMat_discam = np.linalg.inv(rtMat_camdis)
        # rt_dis_inv = np.matmul(rtMat, rtMat_camdis)
        rt_dis_inv = np.matmul(rtMat_camdis, rtMat)
        rt_dis = np.linalg.inv(rt_dis_inv)
        r_dis = rt_dis[0:3, 0:3]
        t_dis = rt_dis[0:3, 3]
        r_dis_inv = rt_dis_inv[0:3, 0:3]
        t_dis_inv = rt_dis_inv[0:3, 3]

        print("rtscreen in world coor is ", rt_dis)
        print("mirror in screen coor is ", rt_dis_inv)
        # reshape t vector
        t_cam = np.reshape(t_cam, (t_cam.shape[0],))
        t_dis = np.reshape(t_dis, (t_dis.shape[0],))
        t_dis_inv = np.reshape(t_dis_inv, (t_dis_inv.shape[0],))
        # reshape for functionality
        t_cam = np.reshape(t_cam, (3, 1))
        t_dis = np.reshape(t_dis, (3, 1))
        t_dis_inv = np.reshape(t_dis_inv, (3, 1))

        # read undistorted image
        img_undistort = cv2.imread(imgUndistortFile)
        gray_undistort = cv2.cvtColor(img_undistort, cv2.COLOR_BGR2GRAY)
        img_undistort_size = gray_undistort.shape
        print(img_undistort_size)

        # cam roor mat:store arrival vectors (vector from camera to the object) in camera coordinate system
        # cam world coor mat: store arrival vectors in world coordinate system
        # camera mirror intersect mat: store intersect points between arrival vectors and mirror in world coordinate system
        # reflec mat: store reflectance vectors (vector from object to screen) in world coordinate system
        # mirror intersect mat trans:store intersect points on mirrors at display coordinate system
        # reflect mat trans: store reflectance vectors in display coordinate system
        # display intersect mat: store display intersect points (intersect points between reflectance vectors and display) on display coordinate system
        # display intersect mat trans: store display intersect points on world coordinate system
        img_coor_mat = np.zeros((3, img_undistort_size[0], img_undistort_size[1]))
        img_coor_mat_rs = np.zeros((3, img_undistort_size[0] * img_undistort_size[1]))
        cam_coor_mat = np.zeros((img_undistort_size[0], img_undistort_size[1], 3))
        cam_coor_mat_rs = np.zeros((3, img_undistort_size[0] * img_undistort_size[1]))
        cam_world_coor_mat = np.zeros((3, img_undistort_size[0] * img_undistort_size[1]))
        cam_mirror_intersect_mat = np.zeros((3, img_undistort_size[0] * img_undistort_size[1]))
        reflect_mat = np.zeros((3, img_undistort_size[0] * img_undistort_size[1]))
        mirror_intersect_mat_trans = np.zeros((3, img_undistort_size[0] * img_undistort_size[1]))
        reflect_mat_trans = np.zeros((3, img_undistort_size[0] * img_undistort_size[1]))
        display_intersect_mat = np.zeros((3, img_undistort_size[0] * img_undistort_size[1]))
        display_intersect_mat_trans = np.zeros((3, img_undistort_size[0] * img_undistort_size[1]))

        print(cam_coor_mat.shape)
        print(np.linalg.inv(cameraMatrix))
        for i in range(cam_coor_mat.shape[0]):
            for j in range(cam_coor_mat.shape[1]):
                image_coor = np.array([j + 1, i + 1, 1])
                img_coor_mat[:, i, j] = image_coor

        img_coor_mat_rs = np.reshape(img_coor_mat, (3, img_undistort_size[0] * img_undistort_size[1]))
        # calculate arrival vector
        cam_coor_mat_rs = np.matmul(np.linalg.inv(cameraMatrix), img_coor_mat_rs)
        cam_world_coor_mat = np.matmul(r_cam, cam_coor_mat_rs)

        # calculate intersect point
        scale_factor1 = -t_cam[2, 0] / cam_world_coor_mat[2, :]
        scale_factor1 = np.reshape(scale_factor1, (1, scale_factor1.shape[0]))
        cam_mirror_intersect_mat = scale_factor1 * cam_world_coor_mat + np.tile(t_cam, (cam_world_coor_mat.shape[1]))
        # calculate reflectance vector in world coor system
        reflect_mat = cam_world_coor_mat - 2 * np.dot(np.array([0, 0, 1]), cam_world_coor_mat) * np.tile(
            np.array([[0], [0], [1]]), (1, cam_world_coor_mat.shape[1]))
        # calculate intersect point in display coor system
        mirror_intersect_mat_trans = np.dot(r_dis_inv, cam_mirror_intersect_mat) + np.tile(t_dis_inv, (
        cam_world_coor_mat.shape[1]))
        # calculate reflectance vector in display coor system
        reflect_mat_trans = np.dot(r_dis_inv, reflect_mat)
        # calculate display intersect point under display coordinate system
        scale_factor2 = -mirror_intersect_mat_trans[2, :] / reflect_mat_trans[2, :]
        display_intersect_mat = scale_factor2 * reflect_mat_trans + mirror_intersect_mat_trans
        # calculate display intersect point under world coordinate system
        display_intersect_mat_trans = np.dot(r_dis, display_intersect_mat) + np.tile(t_dis,
                                                                                     (cam_world_coor_mat.shape[1]))

        # restore the matrix back to the same shape of undistort image
        display_intersect_mat = np.reshape(display_intersect_mat, (3, img_undistort_size[0], img_undistort_size[1]))
        display_intersect_mat_trans = np.reshape(display_intersect_mat_trans,
                                                 (3, img_undistort_size[0], img_undistort_size[1]))
        cam_mirror_intersect_mat = np.reshape(cam_mirror_intersect_mat,
                                              (3, img_undistort_size[0], img_undistort_size[1]))

        ## this part is to display camera, mirror &projector within a unified world coordinate systems

        # reshape t vector
        t_cam = np.reshape(t_cam, (t_cam.shape[0],))
        t_dis = np.reshape(t_dis, (t_dis.shape[0],))
        t_dis_inv = np.reshape(t_dis_inv, (t_dis_inv.shape[0],))

        # mirror corner index
        markerCornerIDs = np.array(
            [np.argwhere(markerIDs == 0), np.argwhere(markerIDs == 2), np.argwhere(markerIDs == 4),
             np.argwhere(markerIDs == 6)])
        markerCornerIDs = markerCornerIDs[:, 0, 0]

        mirror_corner_value = np.array([markerCorners[markerCornerIDs[0]][0][0],
                                        markerCorners[markerCornerIDs[1]][0][3],
                                        markerCorners[markerCornerIDs[2]][0][2],
                                        markerCorners[markerCornerIDs[3]][0][1],
                                        markerCorners[markerCornerIDs[0]][0][0]], dtype=np.int32)
        print("mirror corner index is", mirror_corner_value)
        mirror_corner_world = np.zeros((mirror_corner_value.shape[0], 3))
        for i in range(mirror_corner_world.shape[0]):
            mirror_corner_world[i, :] = cam_mirror_intersect_mat[:, mirror_corner_value[i, 1],
                                        mirror_corner_value[i, 0]]
        print("mirror corner point is", mirror_corner_world)

        # display corners in world coordinate system
        SCALE_FACTOR = self.checker_pixel_size
        half_length_disp = (SCREEN_WIDTH / 2) * SCALE_FACTOR
        half_height_disp = (SCREEN_HEIGHT / 2) * SCALE_FACTOR
        disp_uleft = np.array([-half_length_disp, -half_height_disp, 0])
        disp_lleft = np.array([-half_length_disp, half_height_disp, 0])
        disp_lright = np.array([half_length_disp, half_height_disp, 0])
        disp_uright = np.array([half_length_disp, -half_height_disp, 0])
        disp_uleft_world = np.reshape(np.matmul(r_dis, disp_uleft) + t_dis, (1, t_dis.shape[0]))
        disp_lleft_world = np.reshape(np.matmul(r_dis, disp_lleft) + t_dis, (1, t_dis.shape[0]))
        disp_lright_world = np.reshape(np.matmul(r_dis, disp_lright) + t_dis, (1, t_dis.shape[0]))
        disp_uright_world = np.reshape(np.matmul(r_dis, disp_uright) + t_dis, (1, t_dis.shape[0]))

        disp_corners_world = np.concatenate(
            (disp_uleft_world, disp_lleft_world, disp_lright_world, disp_uright_world, disp_uleft_world), axis=0)
        print("disp corners in world coordinate is ", disp_corners_world)

        # calculate the (0,0) and z axis of camera and display (for verification)
        cam_z1 = np.matmul(r_cam, np.array([0, 0, 1])) + t_cam
        cam_zero = np.matmul(r_cam, np.array([0, 0, 0])) + t_cam
        display_z1 = np.matmul(r_dis, np.array([0, 0, 1])) + t_dis
        display_zero = np.matmul(r_dis, np.array([0, 0, 0])) + t_dis
        # cam_z = np.concatenate((np.reshape(cam_zero, (1, cam_zero.shape[0])), np.reshape(cam_z1, (1, cam_z1.shape[0]))), axis= 0)
        # display_z = np.concatenate((np.reshape(display_zero, (1,display_zero.shape[0])), np.reshape(display_z1, (1, display_z1.shape[0]))), axis= 0)
        cam_z = cam_z1 - cam_zero
        display_z = display_z1 - display_zero
        print('cam z axis in world coor is ', cam_z)
        print("cam zero point in world coor is ", cam_zero)
        print("display z axis in world coor is ", display_z)
        print("display zero in world coor is ", display_zero)
        # plot 
        fig = plt.figure()
        ax = Axes3D(fig)
        ax.plot(mirror_corner_world[:, 0], mirror_corner_world[:, 1], mirror_corner_world[:, 2], 'b-', label='Mirror')
        ax.plot(disp_corners_world[:, 0], disp_corners_world[:, 1], disp_corners_world[:, 2], 'r-', label='Display')
        ax.scatter3D(cam_zero[0], cam_zero[1], cam_zero[2], 'go', label='Camera')
        ax.quiver(cam_zero[0], cam_zero[1], cam_zero[2], cam_z[0], cam_z[1], cam_z[2], length=0.1)
        ax.quiver(display_zero[0], display_zero[1], display_zero[2], display_z[0], display_z[1], display_z[2],
                  length=0.1)

        plt.legend()
        plt.show()
        return display_intersect_mat



from __future__ import print_function
import os, sys, time, math
import subprocess
import numpy as np
import cv2
from PySpinCapture import PySpinCapture as psc
from liveDisplay import ptgCamStream, imgShow, patternDisplay

DEFAULT_DIRECTORY_PATH = ""

DISPLAY_HEIGHT = 1080
DISPLAY_WIDTH = 1920
NUM_PATTERN = 7

#DEFLECTOMETRY_FREQ = 0.9
DEFLECTOMETRY_FREQ = 1.0

class CamControl:
	def __init__(self):

		self.sessionDir = None

		# self.NumPatterns = NUM_PATTERN
		self.displayWidth = DISPLAY_WIDTH
		self.displayHeight = DISPLAY_HEIGHT

		self.setDefPattern()
		self._isMonochrome = True
		self._is16bits = True
		self.Cam = psc(0, self._isMonochrome, self._is16bits)
		self.height = self.Cam.height
		self.width = self.Cam.width

	def setDir(self, directory, sessionName):
		self.directory = directory
		self.sessionName = sessionName
		self.sessionDir = os.path.join(self.directory, self.sessionName)
		print(self.sessionDir)
		if not os.path.exists(self.sessionDir):
			os.makedirs(self.sessionDir)
			
	def setExposure(self, exposure_to_set):
		
		self.Cam.setExposure(exposure_to_set)

	def setGain(self, gain_to_set):

		self.Cam.setGain(gain_to_set)

	def captureImage(self, fname):
		if fname:
			path = os.path.join(self.sessionDir, fname + ".png")
			flag, img = self.Cam.grabFrame()
			if not flag:
				print("[ERROR]: Didn't get the image!!!")
			else:
				cv2.imwrite(path, img, [cv2.IMWRITE_PNG_COMPRESSION, 0])
		else:
			flag, img = self.Cam.grabFrame()
			print("capture!!!!")
			if not flag:
				print("[ERROR]: Didn't get the image!!!")
				return None
			else:
				return img

	def setPatternScale(self, mode=1):
		patternSize = min(self.displayHeight, self.displayWidth)
		boarderSize = math.floor((max(self.displayHeight, self.displayWidth) - patternSize)/2)

		if mode == 0:
			# Img sequences[7]: {gh, ch, gv, cv, g2h, g2v, black}
			self.NumPatterns = 7
		elif mode == 1:
			# Img sequences[8]: {gh, ch, gv, cv, g2h, g2v, black, full}
			self.NumPatterns = 8
		elif mode == 2:
			# Img sequences[9]: {h1, h2, h3, h4, v1, v2, v3, v4, black}
			self.NumPatterns = 9
		elif mode == 3:
			# Img sequences[9]: {h1, h2, h3, h4, v1, v2, v3, v4, black, full}
			self.NumPatterns = 10
		else:
			self.NumPatterns = 8
			print("[WARNING]: Unrecognizable mode. Using Gradient Pattern!!")
		self.patterns = np.zeros((self.displayHeight, self.displayWidth, self.NumPatterns), dtype=np.float32)

		return patternSize, boarderSize

	def setPattern(self):
		patternSize, boarderSize = self.setPatternScale(1)
		# Create spatial coordinates
		x1 = np.linspace(0,1,patternSize)
		y1 = np.linspace(0,1,patternSize)
		x2 = np.linspace(-1,1,patternSize)
		y2 = np.linspace(-1,1,patternSize)

		linearX = x1
		linearY = y1
		linearXinv = 1-x1
		linearYinv = 1-y1
		quadX = (x2**2+1)/2
		quadY = (y2**2+1)/2

		# x direction is flip because of the projection
		self.patterns[:, boarderSize:-boarderSize, 0]= np.tile(linearXinv,(patternSize, 1))
		self.patterns[:, boarderSize:-boarderSize, 1]= np.tile(linearX,(patternSize, 1))
		self.patterns[:, boarderSize:-boarderSize, 2]= np.tile(linearY,(patternSize, 1)).T
		self.patterns[:, boarderSize:-boarderSize, 3]= np.tile(linearYinv,(patternSize, 1)).T
		self.patterns[:, boarderSize:-boarderSize, 4]= np.tile(quadX,(patternSize, 1))
		self.patterns[:, boarderSize:-boarderSize, 5]= np.tile(quadY,(patternSize, 1)).T
		self.patterns[:, boarderSize:-boarderSize, 7] = np.ones((patternSize, patternSize))

	def sinePattern(self, x, y, nu_x, nu_y):
		# Phase Shifts (first entry is white light modulation)
		theta = [0, np.pi / 2, np.pi, 3 / 2 * np.pi]
		[X, Y, Theta] = np.meshgrid(x, y, theta)

		# Calculate Phase Shifts
		phase = (nu_x * X + nu_y * Y) + Theta
		# Simple formula to create fringes between 0 and 1:
		pattern = (np.sin(phase) + 1) / 2

		return pattern

	def setDefPattern(self):
		#patternSize, boarderSize = self.setPatternScale(2)
		patternSize, boarderSize = self.setPatternScale(2)

		patternSize = max(self.displayHeight, self.displayWidth)
		#patternSize = min(self.displayHeight, self.displayWidth)
		
		# Create spatial coordinates
		x = np.linspace(1, self.displayWidth, self.displayWidth)
		y = np.linspace(1, self.displayHeight, self.displayHeight)

		# Frequencies in x and y direction.
		nu_x =  DEFLECTOMETRY_FREQ * 2 * np.pi / patternSize
		nu_y = DEFLECTOMETRY_FREQ * 2 * np.pi / patternSize

		self.patterns[..., 0:4] = self.sinePattern(x, y, nu_x, 0)
		self.patterns[..., 4:8] = self.sinePattern(x, y, 0, nu_y)
		self.patterns[..., 8] = 127 * np.ones((DISPLAY_HEIGHT,DISPLAY_WIDTH))

	def captureSeqImages(self):
		window_name = 'projector'
		cv2.namedWindow(window_name, cv2.WND_PROP_FULLSCREEN)

		# If projector is placed right to main screen (see windows properties in your operating system)
		# if the pattern is displayed at the wrong monitor you need to play around with the coordinates here until the image is displayed at the right screen
		cv2.moveWindow(window_name, self.displayWidth, 0)

		cv2.setWindowProperty(window_name, cv2.WND_PROP_FULLSCREEN,cv2.WINDOW_FULLSCREEN)
		cv2.imshow(window_name, self.patterns[..., 0].astype(np.float32))

		if self._isMonochrome:
			if self._is16bits:
				imgs = np.zeros((self.NumPatterns, self.height, self.width), dtype=np.uint16)
			else:
				imgs = np.zeros((self.NumPatterns, self.height, self.width), dtype=np.uint8)
		else:
			if self._is16bits:
				imgs = np.zeros((self.NumPatterns, self.height, self.width, 3), dtype=np.uint16)
			else:
				imgs = np.zeros((self.NumPatterns, self.height, self.width, 3), dtype=np.uint8)

		cv2.waitKey(200)
		# Loop through 
		for i in range(self.NumPatterns):
			if not i==0:
				cur_img = self.patterns[... ,i]
				cv2.imshow(window_name, cur_img.astype(np.float32))

			cv2.waitKey(400)  # ms # delay time
			# wait for 1000 ms ( syncrhonize this with your aquisition time)
			imgs[i, ...] = self.captureImage(fname=None)

			cv2.waitKey(400) # ms # delay time

		# Don't forgot to close all windows at the end of aquisition
		cv2.destroyAllWindows()

		for i in range(self.NumPatterns):
			fname = os.path.join(self.sessionDir, str(i) + ".png")
			if self._isMonochrome:
				cv2.imwrite(fname, imgs[i, ...], [cv2.IMWRITE_PNG_COMPRESSION, 0])  # RGB to BGR
			else:
				cv2.imwrite(fname, imgs[i, ...][..., ::-1], [cv2.IMWRITE_PNG_COMPRESSION, 0]) #RGB to BGR

	def liveView(self):

		print("[INFO] starting live view thread...")
		viewGrabber = ptgCamStream(self.Cam).start()
		viewShower = imgShow(frame=viewGrabber.read()).start()

		time.sleep(0.1)
		while viewGrabber.running():
			if viewGrabber.stopped or viewShower.stopped:
				viewShower.stop()
				viewGrabber.stop()
				break

			frame = viewGrabber.read()
			viewShower.frame = frame

		viewGrabber.stop()
		viewShower.stop()
		viewGrabber.__exit__()
		viewShower.__exit__()

		viewGrabber.thread.join()
		viewShower.thread.join()
		del viewGrabber
		del viewShower
		cv2.destroyWindow("Live View")

	def setupHDR(self):
		window_name = 'projector'
		cv2.namedWindow(window_name, cv2.WND_PROP_FULLSCREEN)

		# If projector is placed right to main screen (see windows properties in your operating system)
		# if the pattern is displayed at the wrong monitor you need to play around with the coordinates here until the image is displayed at the right screen
		cv2.moveWindow(window_name, self.displayWidth, 0)
		pattern = np.ones((self.displayHeight, self.displayWidth), dtype=np.float32)
		cv2.setWindowProperty(window_name, cv2.WND_PROP_FULLSCREEN, cv2.WINDOW_FULLSCREEN)
		cv2.imshow(window_name, pattern.astype(np.float32))
		cv2.waitKey(200)
		self.Cam.setupHDR()
		cv2.destroyAllWindows()

	def captureHDRSequence(self):
		# TODO: hdr capture with different pattern
		window_name = 'projector'
		cv2.namedWindow(window_name, cv2.WND_PROP_FULLSCREEN)

		# If projector is placed right to main screen (see windows properties in your operating system)
		# if the pattern is displayed at the wrong monitor you need to play around with the coordinates here until the image is displayed at the right screen
		cv2.moveWindow(window_name, self.displayWidth, 0)

		cv2.setWindowProperty(window_name, cv2.WND_PROP_FULLSCREEN, cv2.WINDOW_FULLSCREEN)
		cv2.imshow(window_name, self.patterns[..., 0].astype(np.float32))

		if self._isMonochrome:
			imgs = np.zeros((self.NumPatterns, self.height, self.width), dtype=np.uint8)
		else:
			imgs = np.zeros((self.NumPatterns, self.height, self.width, 3), dtype=np.uint8)

		cv2.waitKey(200)
		# Loop through
		for i in range(self.NumPatterns):
			if not i == 0:
				cur_img = self.patterns[..., i]
				cv2.imshow(window_name, cur_img.astype(np.float32))
			cv2.waitKey(4)  # ms
			# wait for 1000 ms ( syncrhonize this with your aquisition time)
			imgs[i, ...] = self.captureImage(fname=None)
		# cv2.waitKey(5) # ms

		# Don't forgot to close all windows at the end of aquisition
		cv2.destroyAllWindows()

		for i in range(self.NumPatterns):
			fname = os.path.join(self.sessionDir, str(i) + ".png")
			if self._isMonochrome:
				cv2.imwrite(fname, imgs[i, ...], [cv2.IMWRITE_PNG_COMPRESSION, 0])  # RGB to BGR
			else:
				cv2.imwrite(fname, imgs[i, ...][..., ::-1], [cv2.IMWRITE_PNG_COMPRESSION, 0])  # RGB to BGR

	def displayCalib(self):
		# TODO: Need to use threading so that the GUI won't stuck!
		pattern = cv2.imread("./8_24_checker.png", cv2.IMREAD_GRAYSCALE)
		# window_name = 'projector'
		# cv2.namedWindow(window_name, cv2.WND_PROP_FULLSCREEN)
		#
		# # If projector is placed right to main screen (see windows properties in your operating system)
		# # if the pattern is displayed at the wrong monitor you need to play around with the coordinates here until the image is displayed at the right screen
		# cv2.moveWindow(window_name, self.displayWidth, 0)
		#
		# cv2.setWindowProperty(window_name, cv2.WND_PROP_FULLSCREEN, cv2.WINDOW_FULLSCREEN)
		# cv2.imshow(window_name, pattern.astype(np.float32))
		# k = cv2.waitKey(0)
		# if k == 27:  # wait for ESC key to exit
		# 	cv2.destroyAllWindows()

		patternShower = patternDisplay(displayHeight=self.displayHeight, displayWidth=self.displayWidth, pattern=pattern).start()

	def __exit__(self):
		self.Cam.__exit__()




from abc import ABC
from GhostScan.CalibrationSession import CalibrationSession
from GhostScan.Calibrations import RadiometricCalibration
import os


class RadiometricCalibSession(CalibrationSession, ABC):
    def __init__(self, camera, radio_calib, path='CalibrationImages/Radiometric', exposures=0):
        # Set camera, destination path
        self.camera = camera
        self.path = path
        self.g = None
        self.radio_calib = radio_calib
        if exposures == 0:
            # Dark Inside:
            #self.exposures = [500, 750, 1000, 1500, 2000, 4000, 6000, 8000, 10000, 12000, 14000, 16000, 18000, 20000, 22000]
            self.exposures = [100, 200, 400, 600, 1000, 1500, 2000, 4000, 6000, 8000, 10000, 14000, 19000,
                             24000, 31000, 39000, 49000, 60000]
        else:
            self.exposures = exposures

    def capture(self):
        # For radiometric calibration a series of differently exposed images of the same object is required
        if not os.path.exists('CalibrationImages/Radiometric/'):
            if not os.path.exists('CalibrationImages'):
                os.mkdir('CalibrationImages')
            os.mkdir('CalibrationImages/Radiometric/')
        if not os.path.exists('CapturedImages/Radiometric/'):
            if not os.path.exists('CapturedImages'):
                os.mkdir('CapturedImages')
            os.mkdir('CapturedImages/Radiometric/')
        if not os.path.exists('CalibrationNumpyData/Radiometric/'):
            if not os.path.exists('CalibrationNumpyData'):
                os.mkdir('CalibrationNumpyData')
            os.mkdir('CalibrationNumpyData/Radiometric/')
        if not os.path.exists('CapturedNumpyData/Radiometric/'):
            if not os.path.exists('CapturedNumpyData'):
                os.mkdir('CapturedNumpyData')
            os.mkdir('CapturedNumpyData/Radiometric/')
            
        for exp in self.exposures:
            self.camera.setExposure(exp)
            self.camera.getImage(name='Radiometric/'+str(exp),calibration=True)
        print("Capture complete! Radiometric Calibration image sequence saved at: CalibrationImages/Radiometric")

    def calibrate_HDR(self, smoothness=1000):
        # Call radiometric calibration
        # This computes and returns the camera response function and calculates a HDR image saved in path as PNG and .np
        self.g, le = self.radio_calib.get_camera_response(smoothness)
        self.radio_calib.plotCurve("Camera response")
        return self.g, le

    def load_calibration(self):
        self.radio_calib.load_calibration_data()
        self.radio_calib.plotCurve("Camera response")

    def calibrate_image(self, exposure, path='CalibrationImages/Distorted'):
        # Calibrate radiometric calibrated images from a single exposure
        # Returns set of undistorted images and corresponding g function
        imgs, g = self.radio_calib.calibrate_image(exposure, path)
        return imgs, g


# self.exposures = [19, 54, 100, 298, 497, 803, 1599, 3198, 6402, 12798, 16003, 18007, 24005, 39993, 79986,
#                  159971, 300114]
# self.exposures = [19, 38, 76, 152, 304, 608, 1216, 2432, 4864, 9728, 19456, 38912, 77824, 155648, 311296,
#    622592, 1245184, 2490368, 4980736]




# this part is to compare generated "ground truth" normal map and calculated normal map in world coordinate system
 # scale factor, used to discretize the intersection points in metric system (from world coordinate system to image coordinate system)
from __future__ import print_function
import os, sys, time, math
import subprocess
import numpy as np
import cv2
import matplotlib.pyplot as plt
import matplotlib
import GhostScan.Reconstructions.generatephase as gp


def normalInWorld(cam_mirror_intersect_mat, normal_map, half_length, half_height):
    SCALE_FACTOR = 2000

    intersect_point = cam_mirror_intersect_mat
    #varMask = np.load(os.path.join(rootImgPath, "results/mask.npz"))
    #varMask = varMask.f.arr_0
    intersect_x = intersect_point[:,:,0]
    intersect_y = intersect_point[:,:,1]
    normal_map_x = normal_map[:,:,0]
    normal_map_y = normal_map[:,:,1]
    normal_map_z = normal_map[:,:,2]

    
    intersect_x = intersect_x + half_length
    intersect_y = intersect_y + half_height
    intersect_x[intersect_x < 0] = 0
    intersect_x[intersect_x > 2 * half_length] = 0
    intersect_y[intersect_y < 0] = 0
    intersect_y[intersect_y > 2 * half_height] = 0

    print(np.max(intersect_x))
    print(np.max(intersect_y))

    #discretize corld intersect point coordinate.
    intersect_x = intersect_x * SCALE_FACTOR
    intersect_y = intersect_y * SCALE_FACTOR
    intersect_x = np.round(intersect_x)
    intersect_y = np.round(intersect_y)


    normal_map_world = np.zeros((int(np.round(2 * half_height * SCALE_FACTOR) + 1), int(np.round(2 * half_length * SCALE_FACTOR) + 1), 3))
    #varMask_world = np.zeros((int(np.round(2 * half_height * SCALE_FACTOR) + 1), int(np.round(2 * half_length * SCALE_FACTOR) + 1)))
    for i in range(intersect_x.shape[0]):
        for j in range(intersect_x.shape[1]):
            normal_map_world[int(intersect_y[i,j]), int(intersect_x[i,j]), 0] = normal_map_x[i,j] 
            normal_map_world[int(intersect_y[i,j]), int(intersect_x[i,j]), 1] = normal_map_y[i,j]
            normal_map_world[int(intersect_y[i,j]), int(intersect_x[i,j]), 2] = normal_map_z[i,j]
            #varMask_world[int(intersect_y[i,j]), int(intersect_x[i,j])] = varMask[i,j]

    # use mask to filter invalid normals
    #ground_truth_x = ground_truth_x * varMask_world
    #ground_truth_y = ground_truth_y * varMask_world
    #ground_truth_z = ground_truth_z * varMask_world + (varMask_world - 1)
    for i in range(normal_map_world.shape[2]):
        if i == 2:
            normal_map_world[:,:,i] = normal_map_world[:,:,i] #* varMask_world + (varMask_world - 1)
        else:
            normal_map_world[:,:,i] = normal_map_world[:,:,i] #* varMask_world

    return normal_map_world
def normalRotate(normal_map_world, tilt= 0):
    normal_map_rs = np.zeros((3, normal_map_world.shape[0], normal_map_world.shape[1]))
    for i in range(normal_map_world.shape[2]):
        normal_map_rs[i,:,:] = normal_map_world[:,:,i]
    normal_map_rs = np.reshape(normal_map_rs, (3, normal_map_world.shape[0] * normal_map_world.shape[1]))
    tilt_rad = tilt * 2 * math.pi / 360
    rMat = np.array([[1, 0, 0],[0, np.cos(tilt_rad), -np.sin(tilt_rad)],[0, np.sin(tilt_rad), np.cos(tilt_rad)]])
    print(rMat)
    normal_map_rs = np.dot(rMat, normal_map_rs)
    normal_map_temp = np.reshape(normal_map_rs,(3, normal_map_world.shape[0], normal_map_world.shape[1]))
    normal_map_tilted = np.zeros((normal_map_world.shape[0], normal_map_world.shape[1], 3))
    for i in range(normal_map_temp.shape[0]):
        normal_map_tilted[:,:,i] = normal_map_temp[i,:,:]
        
    
    norm = matplotlib.colors.Normalize(vmin= -0.2, vmax = 0.2)
    norm2 = matplotlib.colors.Normalize(vmin= -1, vmax = -0.98)
    # visualize tilted normal map
    """
    plt.figure()
    for i in range(3):
        #plt.figure()
        if i == 2:
            plt.subplot(1,3,i+1)
            plt.imshow(normal_map_tilted[:,:,i], norm= norm2)
            plt.title("Tilt Z")
            plt.tick_params(labelsize=16)
        else:
            plt.subplot(1,3,i+1)
            plt.imshow(normal_map_tilted[:,:,i], norm= norm)
            if i == 1:
                plt.title("Tilt Y")
            else:
                plt.title("Tilt X")
            plt.tick_params(labelsize=16)
        cb = plt.colorbar()
        cb.ax.tick_params(labelsize= 16)
    plt.show()
    """
    return normal_map_tilted

def normalCompare(rootImgPath, normal_map_world, half_length, half_height):
    SCALE_FACTOR = 2500
    # load ground truth normal map generated by Matlab
    intersect_point = np.load(os.path.join(rootImgPath, "results/intersect_world.npz"))
    ground_truth_x = np.loadtxt(os.path.join(rootImgPath, "results/normx_world.txt"))
    ground_truth_y = np.loadtxt(os.path.join(rootImgPath, "results/normy_world.txt"))
    ground_truth_z = np.loadtxt(os.path.join(rootImgPath, "results/normz_world.txt"))
    varMask = np.load(os.path.join(rootImgPath, "results/mask.npz"))
    varMask = varMask.f.arr_0
    intersect_x = intersect_point.f.arr_0[:,:,0]
    intersect_y = intersect_point.f.arr_0[:,:,1]

    intersect_x = intersect_x + half_length
    intersect_y = intersect_y + half_height
    intersect_x[intersect_x < 0] = 0
    intersect_x[intersect_x > 2 * half_length] = 0
    intersect_y[intersect_y < 0] = 0
    intersect_y[intersect_y > 2 * half_height] = 0

    print(np.max(intersect_x))
    print(np.max(intersect_y))

    #discretize corld intersect point coordinate.
    intersect_x = intersect_x * SCALE_FACTOR
    intersect_y = intersect_y * SCALE_FACTOR
    intersect_x = np.round(intersect_x)
    intersect_y = np.round(intersect_y)

    varMask_world = np.zeros((int(np.round(2 * half_height * SCALE_FACTOR) + 1), int(np.round(2 * half_length * SCALE_FACTOR) + 1)))
    for i in range(intersect_x.shape[0]):
        for j in range(intersect_x.shape[1]):
            varMask_world[int(intersect_y[i,j]), int(intersect_x[i,j])] = varMask[i,j]

    norm = matplotlib.colors.Normalize(vmin= -0.2, vmax = 0.2)
    norm2 = matplotlib.colors.Normalize(vmin= -1, vmax = -0.98)
    plt.figure()
    plt.subplot(1,3,1)
    plt.imshow(ground_truth_x, norm= norm)
    plt.title("Ground Truth X")
    cb1 = plt.colorbar()
    plt.tick_params(labelsize=16)
    cb1.ax.tick_params(labelsize= 16)
    plt.subplot(1,3,2)
    plt.imshow(ground_truth_y, norm= norm)
    plt.title("Ground Truth Y")
    cb2 = plt.colorbar()
    plt.tick_params(labelsize=16)
    cb2.ax.tick_params(labelsize= 16)
    plt.subplot(1,3,3)
    plt.imshow(ground_truth_z, norm= norm2)
    plt.title('Ground Truth Z')
    cb3 = plt.colorbar()
    plt.tick_params(labelsize=16)
    cb3.ax.tick_params(labelsize= 16)
    plt.show()
    # tilt normal map to compensate error
    normal_map_world = normalRotate(normal_map_world, tilt=1)
    # calculate errors
    error_x = np.abs(ground_truth_x - normal_map_world[:,:,0]) * varMask_world
    error_y = np.abs(ground_truth_y - normal_map_world[:,:,1]) * varMask_world
    error_z = np.abs(ground_truth_z - normal_map_world[:,:,2]) * varMask_world 
    # try to erode edge errors
    kernel = np.ones((2,2))
    error_x = cv2.erode(error_x, kernel, iterations= 1)
    error_y = cv2.erode(error_y, kernel, iterations= 1)
    error_z = cv2.erode(error_z, kernel, iterations= 1)
    norm3 = matplotlib.colors.Normalize(vmin= 0, vmax = 0.1)
    norm4 = matplotlib.colors.Normalize(vmin= 0, vmax = 0.005)
    plt.figure()
    plt.subplot(1,3,1)
    plt.imshow(error_x, norm= norm3)
    plt.title('Error X')
    cb4 = plt.colorbar()
    plt.tick_params(labelsize=16)
    cb4.ax.tick_params(labelsize= 16)
    plt.subplot(1,3,2)
    plt.imshow(error_y, norm= norm3)
    plt.title('Error Y')
    cb5 = plt.colorbar()
    plt.tick_params(labelsize=16)
    cb5.ax.tick_params(labelsize= 16)
    plt.subplot(1,3,3)
    plt.imshow(error_z, norm= norm4)
    plt.title('Error Z')
    cb6 = plt.colorbar()
    plt.tick_params(labelsize=16)
    cb6.ax.tick_params(labelsize= 16)
    plt.show()

    mean_err_x = np.sum(error_x) / np.sum(varMask_world)
    mean_err_y = np.sum(error_y) / np.sum(varMask_world)
    mean_err_z = np.sum(np.abs(ground_truth_z - normal_map_world[:,:,2]) * varMask_world) / np.sum(varMask_world)
    print("mean error on x, y, z axis is ", mean_err_x, mean_err_y, mean_err_z)
    # save the calculate non-zero error
    error_x = error_x[error_x != 0]
    error_y = error_y[error_y != 0]
    error_z = error_z[error_z != 0]
    np.savetxt(os.path.join(rootImgPath, "results/error_x.txt"), error_x)
    np.savetxt(os.path.join(rootImgPath, "results/error_y.txt"), error_y)
    np.savetxt(os.path.join(rootImgPath, "results/error_z.txt"), error_z)

def calAngFeature(normal_map_world):
    horizontal = np.sqrt(normal_map_world[:,:,0] ** 2 + normal_map_world[:,:,1] ** 2)
    sine_value = horizontal / np.ones((normal_map_world.shape[0], normal_map_world.shape[1]))
    angle_rad = np.arcsin(sine_value)
    angle_value = np.degrees(angle_rad)
    norm = matplotlib.colors.Normalize(vmin= 0, vmax = 20)
    plt.figure()
    plt.imshow(angle_value, norm= norm)
    plt.colorbar()
    plt.title('Angle Descriptors')
    plt.show()




